import argparse
import logging

from pyschism.cmd import common


class BootstrapCli:

    def __init__(self, args: argparse.Namespace):
        self.args = args

    @staticmethod
    def add_subparser_action(subparsers):
        add_bootstrap_options_to_parser(subparsers.add_parser('bootstrap'))

    @property
    def args(self):
        return self._args

    @args.setter
    def args(self, args: argparse.Namespace):
        assert isinstance(args, argparse.Namespace)
        self._args = args


def add_bootstrap_options_to_parser(parser):
    parser.add_argument(
        'project_directory',
        help='Directory where project files will be written to.',
    )
    parser.add_argument(
        "hgrid",
        help='Path to the SCHISM hgrid file.'
    )
    parser.add_argument(
        '--hgrid-crs',
        help='Coordinate reference system string of hgrid.'
    )
    parser.add_argument(
        "--vgrid",
    )
    parser.add_argument(
        "--log-level",
        choices=[name.lower() for name in logging._nameToLevel],
        default="warning",
    )
    parser.add_argument('--overwrite', action='store_true')
    parser.add_argument(
        '--tides',
        '--tidal-database',
        choices=['hamtide', 'tpxo'],
        default='hamtide',
        dest='tidal_database'
    )
    parser.add_argument(
        '--hycom',
        '--baroclinic-database',
        choices=['rtofs', 'gofs'],
        dest='baroclinic_database',
        default='gofs',
    )
    parser.add_argument(
        '--include-velocity',
        action='store_true'
    )
    parser.add_argument('--Z0', type=float)
    parser.add_argument("--cutoff-depth", type=float, default=50.0)
    #common.add_ibctype_to_parser(parser)



import json
import numpy as np
 
#https://java2blog.com/object-of-type-int64-is-not-json-serializable/
class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return super(NpEncoder, self).default(obj)



from typing import List, Union

import numpy as np
import geopandas as gpd
from shapely.geometry import LineString

# from pyschism.forcing.bctides.mod3d import TEM_3D, SAL_3D
# from pyschism.forcing.bctides.nudge import TEM_Nudge, SAL_Nudge
# from pyschism.forcing.bctides.elev2d import Elev2D
# from pyschism.forcing.bctides.uv3d import UV3D
# from pyschism.forcing.bctides.iettype import Iettype
# from pyschism.forcing.bctides.ifltype import Ifltype
# from pyschism.forcing.bctides.itetype import Itetype
# from pyschism.forcing.bctides.isatype import Isatype
# from pyschism.forcing.bctides.itrtype import Itrtype


class Boundaries:
    def __init__(self, hgrid, boundaries: Union[dict, None]):

        ocean_boundaries = []
        land_boundaries = []
        interior_boundaries = []
        if boundaries is not None:
            for ibtype, bnds in boundaries.items():
                if ibtype is None:
                    for id, data in bnds.items():
                        indexes = list(
                            map(hgrid.nodes.get_index_by_id, data["indexes"])
                        )
                        ocean_boundaries.append(
                            {
                                "id": str(id + 1),  # hacking it
                                "index_id": data["indexes"],
                                "indexes": indexes,
                                "geometry": LineString(hgrid.vertices[indexes]),
                                # "iettype": None,
                                # "ifltype": None,
                                # "itetype": None,
                                # "isatype": None,
                                # "itrtype": {},
                                # 'nudge_temperature': None,
                                # 'nudge_salinity': None,
                            }
                        )

                elif str(ibtype).endswith("1"):
                    for id, data in bnds.items():
                        indexes = list(
                            map(hgrid.nodes.get_index_by_id, data["indexes"])
                        )
                        interior_boundaries.append(
                            {
                                "id": str(id + 1),
                                "ibtype": ibtype,
                                "index_id": data["indexes"],
                                "indexes": indexes,
                                "geometry": LineString(hgrid.vertices[indexes]),
                            }
                        )
                else:
                    for id, data in bnds.items():
                        _indexes = np.array(data["indexes"])
                        if _indexes.ndim > 1:
                            # ndim > 1 implies we're dealing with an ADCIRC
                            # mesh that includes boundary pairs, such as weir
                            new_indexes = []
                            for i, line in enumerate(_indexes.T):
                                if i % 2 != 0:
                                    new_indexes.extend(np.flip(line))
                                else:
                                    new_indexes.extend(line)
                            _indexes = np.array(new_indexes).flatten()
                        else:
                            _indexes = _indexes.flatten()
                        indexes = list(map(hgrid.nodes.get_index_by_id, _indexes))

                        land_boundaries.append(
                            {
                                "id": str(id + 1),
                                "ibtype": ibtype,
                                "index_id": data["indexes"],
                                "indexes": indexes,
                                "geometry": LineString(hgrid.vertices[indexes]),
                            }
                        )

        self.open = gpd.GeoDataFrame(
            ocean_boundaries, crs=hgrid.crs if len(ocean_boundaries) > 0 else None
        )
        self.land = gpd.GeoDataFrame(
            land_boundaries, crs=hgrid.crs if len(land_boundaries) > 0 else None
        )
        self.interior = gpd.GeoDataFrame(
            interior_boundaries, crs=hgrid.crs if len(interior_boundaries) > 0 else None
        )
        self.hgrid = hgrid
        self.data = boundaries

    # def elev2d(self):
    #     return Elev2D(self.hgrid)

    # def uv3d(self, vgrid):
    #     return UV3D(self.hgrid, vgrid)

    # def tem3d(self, vgrid):
    #     return TEM_3D(self.hgrid, vgrid)

    # def sal3d(self, vgrid):
    #     return SAL_3D(self.hgrid, vgrid)

    # def TEM_nudge(self, vgrid, data_source, rlmax=1.5, rnu_day=0.25):
    #     return TEM_Nudge(self.hgrid, vgrid, data_source, rlmax, rnu_day)

    # def SAL_nudge(self, vgrid, data_source, rlmax=1.5, rnu_day=0.25):
    #     return SAL_Nudge(self.hgrid, vgrid, data_source, rlmax, rnu_day)

    # def set_forcing(
    #         self,
    #         boundary_id: int,
    #         iettype: Iettype = None,
    #         ifltype: Ifltype = None,
    #         itetype: Itetype = None,
    #         isatype: Isatype = None,
    #         itrtype: Union[Itrtype, List[Itrtype]] = None,
    # ):

    #     def check_input_type(argname, obj, cls):
    #         if not isinstance(obj, cls):
    #             raise TypeError(
    #                 f'Argument {argname} must be of type {cls}, not '
    #                 f'type {type(obj)}.')

    #     def modify_dataframe(argname, obj, cls):
    #         if obj is not None:
    #             check_input_type(argname, obj, cls)
    #             idxs = self.open[self.open['id'] == boundary_id].index.values
    #             for idx in idxs:
    #                 self.open.at[idx, argname] = obj

    #     modify_dataframe('iettype', iettype, Iettype)
    #     modify_dataframe('ifltype', ifltype, Ifltype)
    #     modify_dataframe('itetype', itetype, Itetype)
    #     modify_dataframe('isatype', isatype, Isatype)
    #     modify_dataframe('itrtype', itrtype, Itrtype)


#     def auto_generate(
#         self,
#         threshold=0.,
#         land_ibtype=0,
#         interior_ibtype=1,
#     ):
#         values = self._obj.values
#         if np.any(np.isnan(values)):
#             raise Exception("Mesh contains invalid values. Raster values must"
#                             "be interpolated to the mesh before generating "
#                             "boundaries.")

#         # drop current boundaries
#         self._obj.__dict__['boundaries'] = {}
#         self._obj.__dict__['boundaries'][None] = {}

#         # generate exterior boundaries
#         for ring in self._obj.outer_ring_collection.values():
#             # find boundary edges
#             edge_tag = np.full(ring.shape, 0)
#             edge_tag[np.where(values[ring[:, 0]] < threshold)[0], 0] = -1
#             edge_tag[np.where(values[ring[:, 1]] < threshold)[0], 1] = -1
#             edge_tag[np.where(values[ring[:, 0]] >= threshold)[0], 0] = 1
#             edge_tag[np.where(values[ring[:, 1]] >= threshold)[0], 1] = 1
#             # sort boundary edges
#             ocean_boundary = list()
#             land_boundary = list()
#             for i, (e0, e1) in enumerate(edge_tag):
#                 if np.any(np.asarray((e0, e1)) == -1):
#                     ocean_boundary.append(tuple(ring[i, :]))
#                 elif np.any(np.asarray((e0, e1)) == 1):
#                     land_boundary.append(tuple(ring[i, :]))
#             ocean_boundaries = sort_edges(ocean_boundary)
#             land_boundaries = sort_edges(land_boundary)
#             _bnd_id = len(self._obj.boundaries[None])
#             for bnd in ocean_boundaries:
#                 e0, e1 = [list(t) for t in zip(*bnd)]
#                 e0 = list(map(self._obj.get_vertex_id_by_index, e0))
#                 data = e0 + [self._obj.get_vertex_id_by_index(e1[-1])]
#                 self._set_boundary_data(self._obj.vertex_id,
#                                         None, _bnd_id, data)
#                 _bnd_id += 1
#             # add land boundaries
#             if land_ibtype not in self._obj.boundaries:
#                 self._add_boundary_type(land_ibtype)
#             _bnd_id = len(self._obj.boundaries[land_ibtype])
#             for bnd in land_boundaries:
#                 e0, e1 = [list(t) for t in zip(*bnd)]
#                 e0 = list(map(self._obj.get_vertex_id_by_index, e0))
#                 data = e0 + [self._obj.get_vertex_id_by_index(e1[-1])]
#                 self._set_boundary_data(self._obj.vertex_id, land_ibtype,
#                                         _bnd_id, data)
#                 _bnd_id += 1
#         # generate interior boundaries
#         _bnd_id = 0
#         _interior_boundaries = defaultdict()
#         for interiors in self._obj.inner_ring_collection.values():
#             for interior in interiors:
#                 e0, e1 = [list(t) for t in zip(*interior)]
#                 if signed_polygon_area(self._obj._vertices[e0, :]) < 0:
#                     e0 = list(reversed(e0))
#                     e1 = list(reversed(e1))
#                 e0 = list(map(self._obj.get_vertex_id_by_index, e0))
#                 e0.append(e0[0])
#                 _interior_boundaries[_bnd_id] = e0
#                 _bnd_id += 1
#         self._add_boundary_type(interior_ibtype)
#         for bnd_id, data in _interior_boundaries.items():
#             self._set_boundary_data(self._obj._vertex_id, interior_ibtype,
#                                     bnd_id, data)

#     @figure
#     def plot(
#         self,
#         ibtype,
#         id,
#         tags=True,
#         axes=None,
#         show=False,
#         figsize=None,
#         **kwargs
#     ):
#         boundary = list(map(self._obj.get_vertex_index_by_id,
#                             self._obj.boundaries[ibtype][id].indexes))
#         p = axes.plot(self._obj.x[boundary], self._obj.y[boundary], **kwargs)
#         if tags:
#             axes.text(
#                 self._obj.x[boundary[len(boundary)//2]],
#                 self._obj.y[boundary[len(boundary)//2]],
#                 f"ibtype={ibtype}\nid={id}",
#                 color=p[-1].get_color()
#                 )
#         return axes

#     @figure
#     def make_plot(
#         self,
#         axes=None,
#         show=False,
#         figsize=None,
#         **kwargs
#     ):
#         kwargs.update({'axes': axes})
#         for ibtype, bnds in self._obj.boundaries.items():
#             for id in bnds:
#                 axes = self.plot(ibtype, id, **kwargs)
#                 kwargs.update({'axes': axes})
#         return kwargs['axes']



from pyschism.mesh.hgrid import Hgrid
from pyschism.mesh.fgrid import DragCoefficient

hgrid=Hgrid.open('./hgrid.gr3', crs='epsg:4326')
depth1=-1.0
depth2=-3.0
bfric_river=0.0025
bfric_land=0.025
fgrid=DragCoefficient.linear_with_depth(hgrid, depth1, depth2, bfric_river, bfric_land)

#modify values in regions
regions=['GoME+0.001.reg', 'Lake_Charles_0.reg']
values=[0.001, 0.0]
flags=[1, 0] # 0: reset to given value, 1: add given value to the current value
for reg, value, flag in zip(regions, values, flags):
    fgrid.modify_by_region(hgrid, f'./{reg}', value, depth1, flag)

fgrid.write('drag.gr3', overwrite=True)



from pyschism.mesh import Hgrid
from pyschism.mesh.gridgr3 import Shapiro

if __name__ == '__main__':
    hgrid = Hgrid.open('./hgrid.gr3', crs='EPSG:4326')

    shapiro_max = 0.5
    threshold_slope = 0.5
    depths = [-99999, 20, 50]  # tweaks in shallow waters
    shapiro_vals1 = [0.2, 0.2, 0.05]  # tweaks in shallow waters
    regions = [
        f'./coastal_0.2.cpp.reg',
        f'./coastal_0.5_1.cpp.reg',
        f'./coastal_0.5_2.cpp.reg'
    ]  # tweaks in regions, the order matters
    shapiro_vals2 = [0.2, 0.5, 0.5]  # tweaks in regions, the order matters
    i_set_add_s = [0, 0, 0]

    shapiro = Shapiro.slope_filter(
        hgrid, shapiro_vals1, depths, shapiro_max, threshold_slope,
        regions, shapiro_vals2, i_set_add_s, lonc=-77.07, latc=24.0)
    shapiro.write('shapiro.gr3', overwrite=True)



from datetime import datetime, timedelta
import logging

import pathlib
from typing import Union

import f90nml

# import pytz

from pyschism import dates
from pyschism.mesh.fgrid import NchiType


logger = logging.getLogger(__name__)


# class FlagIcDescriptor:

#     ic_types = [
#         # (name, type, index)
#         ("ic_temp", gridgr3.TempIc, 0),
# ]

#     def __init__(self, name, ic_type, index):
#         self.type = ic_type
#         self.name = name
#         self.index = index

#     def __get__(self, obj, val):
#         return bool(obj.flag_ic[self.index])

#     def __set__(self, obj, val: bool):
#         if not isinstance(val, bool):
#             raise TypeError(
#                 f"Argument to {self.name} must be boolean, not " f"type {type(val)}."
#             )
#         obj.flag_ic[self.index] = int(val)


class OptMeta(type):
    def __new__(meta, name, bases, attrs):

        PARAM_DEFAULTS = f90nml.read(pathlib.Path(__file__).parent / 'param.nml')["opt"]
        for key, value in PARAM_DEFAULTS.items():
            if key not in attrs:
                if isinstance(value, list):
                    attrs[key] = len(value) * [0]
                else:
                    attrs[key] = None
        # for ic_type in FlagIcDescriptor.ic_types:
        #     attrs[ic_type] = FlagIcDescriptor()
        return type(name, bases, attrs)


class OPT(metaclass=OptMeta):
    """Provides error checking implementation for OPT group"""

    def __init__(
        self,
        start_date: datetime = None,
        dramp: Union[int, float, timedelta] = None,
        drampbc: Union[int, float, timedelta] = None,
        dramp_ss: Union[float, timedelta] = None,
        drampwafo: Union[float, timedelta] = None,
        drampwind: Union[float, timedelta] = None,
        ics=None,
        nws=None,
        nchi: Union[int, NchiType] = None,
        hmin_man=None,
        dbz_min=None,
        dbz_decay=None,
        ic_elev=None,
    ):
        self.start_date = start_date
        self.dramp = dramp
        self.drampbc = drampbc
        self.dramp_ss = dramp_ss
        self.drampwafo = drampwafo
        self.drampwind = drampwind
        self.ics = ics
        self.nchi = nchi
        self.hmin_man = hmin_man
        self.dbz_min = dbz_min
        self.dbz_decay = dbz_decay
        self.ic_elev = ic_elev
        self.nws = nws

    def __str__(self):
        data = []
        PARAM_DEFAULTS = f90nml.read(pathlib.Path(__file__).parent / 'param.nml')["opt"]
        for key, _ in PARAM_DEFAULTS.items():
            current = getattr(self, key)
            if current is not None:
                if not isinstance(current, list):
                    data.append(f"  {key}={current}")
                else:
                    for i, state in enumerate(current):
                        if state:
                            current.append(f"  {current}({i+1}) = 1")
        data = "\n".join(data)
        return f"&OPT\n{data}\n/"

    def to_dict(self):
        data = {}
        PARAM_DEFAULTS = f90nml.read(pathlib.Path(__file__).parent / 'param.nml')["opt"]
        for key, _ in PARAM_DEFAULTS.items():
            current = getattr(self, key)
            if current is not None:
                if not isinstance(current, list):
                    data[key] = current
                else:
                    data[key] = len(current) * [0]
                    for i, state in enumerate(current):
                        if state:
                            data[key][i] = 1
        return data

    @property
    def start_date(self) -> Union[datetime, None]:
        return self._start_date

    @start_date.setter
    def start_date(self, start_date: Union[datetime, None]):
        if start_date is not None:
            start_date = dates.localize_datetime(start_date)
            self.start_year = start_date.year
            self.start_month = start_date.month
            self.start_day = start_date.day
            self.start_hour = start_date.hour
            self.start_hour += start_date.minute / 60.0
            self.utc_start = -start_date.utcoffset().total_seconds() / 3600.0  # type: ignore[union-attr]  # noqa: E501
        else:
            self.start_year = None
            self.start_month = None
            self.start_day = None
            self.start_hour = None
            self.utc_start = None
        self._start_date = start_date

    @property
    def start_year(self) -> Union[int, None]:
        return self._start_year

    @start_year.setter
    def start_year(self, start_year: Union[int, None]):
        if start_year is not None:
            start_year = int(start_year)
        self._start_year = start_year

    @property
    def start_month(self) -> Union[int, None]:
        return self._start_month

    @start_month.setter
    def start_month(self, start_month: Union[int, None]):
        if start_month is not None:
            start_month = int(start_month)
        self._start_month = start_month

    @property
    def start_day(self) -> Union[int, None]:
        return self._start_day

    @start_day.setter
    def start_day(self, start_day: Union[int, None]):
        if start_day is not None:
            start_day = int(start_day)
        self._start_day = start_day

    @property
    def start_hour(self) -> Union[float, None]:
        return self._start_hour

    @start_hour.setter
    def start_hour(self, start_hour: Union[float, None]):
        if start_hour is not None:
            start_hour = float(start_hour)
        self._start_hour = start_hour

    @property
    def utc_start(self) -> Union[float, None]:
        return self._utc_start

    @utc_start.setter
    def utc_start(self, utc_start: Union[float, None]):
        if utc_start is not None:
            utc_start = float(utc_start)
        self._utc_start = utc_start

    @property
    def dramp(self) -> Union[float, None]:
        return self._dramp

    @dramp.setter
    def dramp(self, dramp: Union[float, timedelta, None]):
        if dramp is not None:
            if not isinstance(dramp, timedelta):
                dramp = timedelta(days=float(dramp))
            self._dramp = float(dramp / timedelta(days=1))
        else:
            self._dramp = None

    @property
    def nramp(self) -> Union[int, None]:
        if not hasattr(self, "_nramp") and self.dramp is not None:
            return 1
        elif hasattr(self, "_nramp"):
            return self._nramp

    @nramp.setter
    def nramp(self, nramp: Union[int, None]):
        if nramp not in [0, 1]:
            raise ValueError("Argument nramp must be 0, 1.")
        self._nramp = nramp

    @nramp.deleter
    def nramp(self):
        if hasattr(self, "_nramp"):
            del self._nramp

    @property
    def drampbc(self) -> Union[float, None]:
        return self._drampbc

    @drampbc.setter
    def drampbc(self, drampbc: Union[float, timedelta, None]):
        if drampbc is not None:
            if not isinstance(drampbc, timedelta):
                drampbc = timedelta(days=float(drampbc))
            self._drampbc = float(drampbc / timedelta(days=1))
        else:
            self._drampbc = None

    @property
    def nrampbc(self) -> Union[int, None]:
        if not hasattr(self, "_nrampbc") and self.drampbc is not None:
            return 1
        elif hasattr(self, "_nrampbc"):
            return self._nrampbc

    @nrampbc.setter
    def nrampbc(self, nrampbc: Union[int, None]):
        if nrampbc not in [0, 1]:
            raise ValueError("Argument nrampbc must be 0, 1.")
        self._nrampbc = nrampbc

    @nrampbc.deleter
    def nrampbc(self):
        if hasattr(self, "_nrampbc"):
            del self._nrampbc

    @property
    def dramp_ss(self) -> Union[float, None]:
        return self._dramp_ss

    @dramp_ss.setter
    def dramp_ss(self, dramp_ss: Union[float, timedelta, None]):
        if dramp_ss is not None:
            if not isinstance(dramp_ss, timedelta):
                dramp_ss = timedelta(days=float(dramp_ss))
            self._dramp_ss = float(dramp_ss / timedelta(days=1))
        else:
            self._dramp_ss = None

    @property
    def nramp_ss(self) -> Union[int, None]:
        if not hasattr(self, "_nramp_ss") and self.dramp_ss is not None:
            return 1
        elif hasattr(self, "_nramp_ss"):
            return self._nramp_ss

    @nramp_ss.setter
    def nramp_ss(self, nramp_ss: Union[int, None]):
        if nramp_ss not in [0, 1]:
            raise ValueError("Argument nramp_ss must be 0, 1.")
        self._nramp_ss = nramp_ss

    @nramp_ss.deleter
    def nramp_ss(self):
        if hasattr(self, "_nramp_ss"):
            del self._nramp_ss

    @property
    def drampwafo(self) -> Union[float, None]:
        return self._drampwafo

    @drampwafo.setter
    def drampwafo(self, drampwafo: Union[float, timedelta, None]):
        if drampwafo is not None:
            if not isinstance(drampwafo, timedelta):
                drampwafo = timedelta(days=float(drampwafo))
            self._drampwafo = float(drampwafo / timedelta(days=1))
        else:
            self._drampwafo = None

    @property
    def nrampwafo(self) -> Union[int, None]:
        if not hasattr(self, "_nrampwafo") and self.drampwafo is not None:
            return 1
        elif hasattr(self, "_nrampwafo"):
            return self._nrampwafo

    @nrampwafo.setter
    def nrampwafo(self, nrampwafo: Union[int, None]):
        if nrampwafo not in [0, 1]:
            raise ValueError("Argument nrampwafo must be 0, 1.")
        self._nrampwafo = nrampwafo

    @nrampwafo.deleter
    def nrampwafo(self):
        if hasattr(self, "_nrampwafo"):
            del self._nrampwafo

    @property
    def drampwind(self) -> Union[float, None]:
        return self._drampwind

    @drampwind.setter
    def drampwind(self, drampwind: Union[float, timedelta, None]):
        if drampwind is not None:
            if not isinstance(drampwind, timedelta):
                drampwind = timedelta(days=float(drampwind))
            self._drampwind = float(drampwind / timedelta(days=1))
        else:
            self._drampwind = None

    @property
    def nrampwind(self) -> Union[int, None]:
        if not hasattr(self, "_nrampwind") and self.drampwind is not None:
            return 1
        elif hasattr(self, "_nrampwind"):
            return self._nrampwind

    @nrampwind.setter
    def nrampwind(self, nrampwind: Union[int, None]):
        if nrampwind not in [0, 1]:
            raise ValueError("Argument nrampwind must be 0, 1.")
        self._nrampwind = nrampwind

    @nrampwind.deleter
    def nrampwind(self):
        if hasattr(self, "_nrampwind"):
            del self._nrampwind

    @property
    def nchi(self) -> Union[int, None]:
        return self._nchi

    @nchi.setter
    def nchi(self, nchi: Union[int, NchiType, None]):
        if not isinstance(nchi, NchiType) and nchi is not None:
            nchi = NchiType(nchi).value
        self._nchi = nchi

    @property
    def ic_elev(self) -> None:
        return self._ic_elev

    @ic_elev.setter
    def ic_elev(self, ic_elev: Union[int, None]):
        assert ic_elev in [0, 1, None]
        self._ic_elev = ic_elev

    # @property
    # def flag_ic(self) -> List[int]:
    #     if not hasattr(self, "_flag_ic"):
    #         self._flag_ic = len(FlagIcDescriptor.ic_types) * [0]
    #     return self._flag_ic


# --- I think none of this is needed anymore but kept for here just in case.

# class Dramp:

#     def __set__(self, obj, dramp: Union[int, float, timedelta, None]):

#         if not isinstance(dramp, (int, float, timedelta, type(None))):
#             raise TypeError("Argument drampbc must be an int, float, "
#                             "timedelta, or None.")

#         if isinstance(dramp, (int, float)):
#             dramp = timedelta(days=dramp)

#         if dramp is not None:
#             obj.nramp = 1

#         obj.__dict__['dramp'] = dramp

#     def __get__(self, obj, val):
#         return obj.__dict__.get('dramp')


# class Drampbc:

#     def __set__(self, obj, drampbc: Union[int, float, timedelta, None]):
#         if not isinstance(drampbc, (int, float, timedelta, type(None))):
#             raise TypeError("Argument drampbc must be an int, float, "
#                             "timedelta, or None.")

#         if isinstance(drampbc, (int, float)):
#             drampbc = timedelta(days=drampbc)

#         if drampbc is not None:
#             obj.nrampbc = 1

#         obj.__dict__['drampbc'] = drampbc

#     def __get__(self, obj, val):
#         return obj.__dict__.get('drampbc')


# class StartDate:

#     def __set__(self, obj, start_date: Union[datetime, None]):
#         if not isinstance(start_date, (datetime, type(None))):
#             raise TypeError("Argument start_date must be of type datetime or "
#                             "None. The datetime object will be assumed to be "
#                             "in UTC if ")
#         if start_date is not None:
#             if start_date.tzinfo is None \
#                     or start_date.tzinfo.utcoffset(start_date) is None:
#                 start_date = start_date.replace(tzinfo=pytz.utc)
#             obj.start_year = start_date.year
#             obj.start_month = start_date.month
#             obj.start_day = start_date.day
#             obj.start_hour = start_date.hour
#             obj.start_hour += start_date.minute / 60.
#             obj.utc_start = -start_date.utcoffset().total_seconds() / 3600  # type: ignore[union-attr]  # noqa: E501
#             # get rid of "negative" zero
#             obj.utc_start = +0. if obj.utc_start == -0. \
#                 else obj.utc_start
#         obj.__dict__['start_date'] = start_date

#     def __get__(self, obj, val):
#         return obj.__dict__.get('start_date')


# class Nchi:

#     def __set__(self, obj, fgrid):
#         nchi = fgrid.nchi
#         if nchi == -1:
#             obj.hmin_man = fgrid.hmin_man
#         if obj.nchi == 1:
#             obj.dbz_min = fgrid.dbz_min
#             obj.dbz_decay = fgrid.dbz_decay
#         obj.__dict__['nchi'] = nchi

#     def __get__(self, obj, val):
#         return obj.__dict__.get('nchi')


# class Ics:

#     def __set__(self, obj, ics: int):
#         obj.__dict__['ics'] = ics

#     def __get__(self, obj, val):
#         return obj.__dict__.get('ics')


# class Sfea0:

#     def __set__(self, obj, sfea0: float):
#         obj.__dict__['sfea0'] = sfea0

#     def __get__(self, obj, val):
#         return obj.__dict__.get('sfea0')


# class Ncor:

#     def __set__(self, obj, ncor: Coriolis):
#         if not isinstance(ncor, Coriolis):
#             raise TypeError(f"ncor must be of type {Coriolis}, not type "
#                             f"{type(ncor)}.")
#         obj.__dict__['ncor'] = ncor.value

#     def __get__(self, obj, val):
#         return obj.__dict__.get('ncor')


# class Ihot:

#     def __set__(self, obj, ihot: int):
#         if not isinstance(ihot, int):
#             raise TypeError(f"ihot must be of type {int}, not type "
#                             f"{type(ihot)}.")
#         if ihot not in [0, 1]:
#             raise ValueError("ihot must be 0 or 1")

#         obj.__dict__['ihot'] = ihot

#     def __get__(self, obj, val):
#         return obj.__dict__.get('ihot')


# class Nws:
#     def __set__(self, obj, nws: NWS):
#         if not isinstance(nws, NWS):
#             raise TypeError(
#                 f"nws must be of type {NWS}, not type {type(nws)}.")

#         if obj.start_date is None:
#             raise ValueError(
#                 "Can't initialize atmospheric data without start_date")
#             nws._start_date = obj.start_date

#         if isinstance(nws, NWS2):
#             obj.__dict__['nws'] = 2

#         else:
#             raise NotImplementedError(
#                 f'NWS type {type(nws)} is not implemented.')

#     def __get__(self, obj, val):
#         return obj.__dict__.get('nws')



#! /usr/bin/env python
import pathlib
from datetime import datetime, timedelta, timezone
import logging

from pyschism.mesh import Hgrid, Vgrid, Fgrid
from pyschism import ModelDomain, ModelDriver, Stations
from pyschism.forcing import Tides
from pyschism.forcing.atmosphere.nws.nws2 import NWS2
from pyschism.forcing.atmosphere.gfs import GlobalForecastSystem as GFS
from pyschism.forcing.atmosphere.hrrr import HRRR

# https://eev.ee/blog/2012/05/23/python-faq-descriptors/

PARENT = pathlib.Path(__file__).parent

logging.basicConfig(level=logging.INFO)

_logger = logging.getLogger(__name__)

if __name__ == '__main__':

    hgrid = Hgrid.open(PARENT / 'hgrid.gr3', crs='EPSG:4326')
    vgrid = Vgrid()
    fgrid = Fgrid.open(PARENT / 'manning.gr3', crs='EPSG:4326')

    # setup model domain
    domain = ModelDomain(hgrid, vgrid, fgrid)

    # set tidal boundary conditions
    elevbc = Tides()
    elevbc.use_all()  # activate all forcing constituents

    # connect the boundary condition to the domain
    domain.add_boundary_condition(elevbc)

    domain.set_atmospheric_forcing(NWS2(GFS(), HRRR()))

    # Use int or float for seconds, or timedelta objects for pythonic
    # specifications
    dt = timedelta(seconds=150.)
    rnday = timedelta(days=30.)

    # Use an integer for number of steps or a timedelta to approximate
    # number of steps internally based on timestep
    nspool = timedelta(minutes=30.)

    # The dramp and start_date are optional parameters.
    # start_date is required when using forcings that
    # require some natural dates, (e.g. tides, winds, T&S ...)
    # If using a forcing that requires a start date is invoked, and no start
    # date is provided, the driver will throw an exception.
    # tzinfo is optional, UTC assumed if not provided
    dramp = 0.1 * rnday
    start_date = datetime(2017, 9, 18, 4, 0,
                          tzinfo=timezone(timedelta(hours=-4))
                          ) - dramp

    # Now we add station outputs
    stations = Stations.from_file(
        PARENT / 'station.in',
        timedelta(minutes=6.),
        elev=True,
        u=True,
        v=True
    )

    # init the model driver
    driver = ModelDriver(
        domain,
        dt,
        rnday,
        dramp=dramp,
        start_date=start_date,
        # stations=stations,
        nspool=nspool,
        elev=True,
    )

    # Output requests, as well as most other namelist variables, can be
    # modified post instantiation through their corresponding properties.
    # For example, the following line activates the depth averaged horizontal
    # velocity output request:
    driver.param.schout.dahv = True

    # write files to disk
    outdir = pathlib.Path('staging')
    driver.write(outdir, overwrite=True)



#! /usr/bin/env python
from datetime import datetime, timedelta
import logging
import pathlib
import shutil

from pyschism import dates
from pyschism.driver import ModelConfig
from pyschism.forcing.nws import NWS2, GFS, HRRR
from pyschism.forcing.source_sink import NWM
from pyschism.mesh import Hgrid

SCHISM_BINARY = 'pschism_TVD-VL'


def test_barotropic_forecast():

    test_output_directory = pathlib.Path("/tmp/test")
    if test_output_directory.exists():
        shutil.rmtree(test_output_directory)

    config = ModelConfig(
        Hgrid.open(
            "https://raw.githubusercontent.com/geomesh/test-data/main/NWM/hgrid.ll",
            crs="epsg:4326",
        ),
        flags = [[3, 3, 0, 0]],
        constituents = 'major',
        database="tpxo",
        #nws=NWS2(GFS(), HRRR()),
        source_sink=NWM(
            # aggregation_radius=4000.
        ),
    )

    #config.forcings.nws.sflux_2.inventory.file_interval = timedelta(hours=6)

    # create reference dates
    #nearest_cycle = dates.nearest_cycle()
    start_date = datetime(2023, 9, 10)
    spinup_time = timedelta(days=1)
    coldstart = config.coldstart(
        start_date= start_date,
        end_date=start_date + timedelta(days=3),
        timestep=300.0,
        dramp=spinup_time,
        dramp_ss=spinup_time,
        drampwind=spinup_time,
        nspool=timedelta(hours=1),
        elev=True,
        dahv=True,
    )

    # optionally run or write the coldstart object
    if shutil.which(SCHISM_BINARY) is not None:
        coldstart.run(
            test_output_directory / "coldstart",
            overwrite=True,
        )

    else:
        coldstart.write(test_output_directory / "coldstart", overwrite=True)

    hotstart = config.hotstart(
        coldstart,
        end_date=timedelta(days=1) + timedelta(hours=23.0),
        timestep=300.0,
        nspool=timedelta(hours=1.0),
        elev=True,
        dahv=True,
    )

    if shutil.which(SCHISM_BINARY) is not None:
        hotstart.run(
            test_output_directory / "hotstart",
            overwrite=True,
        )

    else:
        hotstart.write(test_output_directory / "hotstart", overwrite=True)


if __name__ == "__main__":

    logging.basicConfig(
        format="[%(asctime)s] %(name)s %(levelname)s: %(message)s",
        force=True,
    )
    logging.captureWarnings(True)
    logging.getLogger("pyschism").setLevel(logging.DEBUG)
    test_barotropic_forecast()



from datetime import datetime, timezone, timedelta
import pathlib
from typing import Union, List
import warnings

import f90nml
import matplotlib.pyplot as plt
import numpy as np

from pyschism.utils.coops import CoopsDataCollector
from pyschism.enums import StationOutputIndex, StationOutputVariables


class StationsOutput:

    def __init__(self, outputs: Union[str, pathlib.Path],
                 stations_file: Union[str, pathlib.Path] = None,
                 ):
        outputs = pathlib.Path(outputs)
        if not outputs.is_dir():
            raise Exception(f'{str(outputs)} is not a directory.')
        self._station_id: List = []
        if stations_file is None:
            stations_file = outputs.resolve() / '../station.in'
            if not stations_file.is_file():
                stations_file = None
            else:
                with open(stations_file) as f:
                    f.readline()
                    for station in range(int(f.readline())):
                        line = f.readline()
                        if '!' in line:
                            self._station_id.append(line.split('!')[-1])
                        else:
                            self._station_id.append(None)
        self._manifest = list(outputs.glob('staout_*'))

        # find the start_date. Let's first check if it's on the param.nml file
        param_file = outputs.resolve() / '../param.nml'
        if param_file.is_file():
            param = f90nml.read(param_file)
            start_year = param['opt'].get('start_year')
            start_month = param['opt'].get('start_month')
            start_day = param['opt'].get('start_day')
            start_hour = param['opt'].get('start_hour')
            utc_start = param['opt'].get('utc_start')
            if None in [start_year, start_month, start_day, start_hour,
                        utc_start]:
                warnings.warn('Could not determine start date automatically.')
                self._start_date = None
            else:
                self._start_date = datetime(start_year, start_month,
                                            start_day, int(start_hour), 0,
                                            tzinfo=timezone(
                                                timedelta(hours=-utc_start)))
        self._rndays = None
        for file in self._manifest:
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                data = np.loadtxt(file)
                if len(w) != 0:
                    pass
                else:
                    self._rndays = timedelta(seconds=data[-1, 0])
                    break

        if self.rndays is None:
            raise Exception("Ouptut directory doesn't contain any station "
                            "output data.")

    def set_start_date(self, start_date: datetime):
        self._start_date = start_date

    def get_station_id(self, index):
        return self._station_id[index].strip()

    def plot(self, variable, station_index=None, show=False):
        filenames = [path.name for path in self._manifest]
        var_index = StationOutputIndex[
            StationOutputVariables(variable).name].value + 1

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            data = np.loadtxt(
                self._manifest[filenames.index(f'staout_{var_index}')])
            if len(w) != 0:
                raise AttributeError(f'Empty record for variable {variable}.')

        if self.start_date is not None:
            start_date = self.start_date
            dates = [start_date + timedelta(seconds=t) for t in data[:, 0]]
        else:
            dates = data[:, 0]

        coops = CoopsDataCollector()
        if station_index is None:
            for i in range(data.shape[1] - 2):
                plt.figure(i)
                if self._obs:
                    station_id = self.get_station_id(i)
                    if station_id is not None:
                        obs = coops.fetch(
                            station_id, variable, self.start_date,
                            self.rndays)
                        plt.plot(obs['datetime'], obs['values'])
                        plt.title(obs['name'])
                        # fig, ax = plt.subplots(figsize=(8, 3))
                        # import matplotlib.dates as mdates
                        # plt.plot_date(obs['datetime'], obs['values'], ls='solid', lw=1.5, aa=True,marker='None',color='g')
                        # ax = plt.gca()
                        # dayFmt = mdates.DateFormatter('%a-%b-%d')
                        # hrFmt = mdates.DateFormatter('%H:00')
                        # ax.xaxis.set_major_formatter(dayFmt)
                        # ax.xaxis.set_major_locator(mdates.DayLocator())
                        # ax.xaxis.set_minor_locator(mdates.HourLocator(byhour=[12]))
                        # ax.xaxis.set_minor_formatter(hrFmt)
                        # plt.grid(b=True, which='both', linestyle='dotted')
                        # plt.show()
                plt.plot(dates, data[:, i+1])

        else:
            plt.plot(data[:, 0], data[:, station_index])

        if show is True:
            plt.show()
        # import pdb; pdb.set_trace()
        # with open(self.manifest[filenames.index(f'staout_{index}')]) as f:
        #     for line in f:
        #         print(len(line.split()))
        #         exit()

    @property
    def start_date(self):
        return self._start_date

    @property
    def rndays(self):
        return self._rndays


from datetime import datetime, timedelta
import pathlib
import os
from typing import Union
import logging

import numpy as np
import pandas as pd
import xarray as xr
import boto3
from botocore import UNSIGNED
from botocore.config import Config
import fsspec
import zarr
from matplotlib.transforms import Bbox

from pyschism.dates import nearest_cycle

logger = logging.getLogger(__name__)

class AWSZarrInventory:
    '''HRRR AWS Zarr bucket
    Create sflux from HRRr Zarr files.

    Notes
    -------
    DSWRF has missing data:
    https://mesowest.utah.edu/html/hrrr/zarr_documentation/html/zarr_variables.html
    If only need air/prc, this method is faster than generating sflux from grib2 files (hrrr3).
    '''

    def __init__(self, bbox: Union[list, Bbox]=None):
        '''
        Parameters
        -------
        bbox: matplotlib.transforms.Bbox object or a list [xmin, ymin, xmax, ymax]
        '''

        if bbox is None:
            raise ValueError('Bbox is needed!')

        if not isinstance(bbox, Bbox) and not isinstance(bbox, list):
            raise TypeError('bbox is a list or matplotlib.transforms.Bbox object!')

        if isinstance(bbox, list):
            xmin, ymin, xmax, ymax = [v for v in bbox]
            self.bbox =  Bbox.from_extents(xmin, ymin, xmax, ymax)
        else:
            self.bbox = bbox

    def gen_sflux(self, outdir: Union[str, os.PathLike], start_date=None, stack=None, air=True, prc=True, rad=True): 
        self.start_date = nearest_cycle() if start_date is None else start_date
        airVars = {'stmp': ['TMP', '2m_above_ground'],
                   'spfh': ['SPFH', '2m_above_ground'],
                   'uwind': ['UGRD', '10m_above_ground'],
                   'vwind': ['VGRD', '10m_above_ground'],
                   'prmsl': ['MSLMA', 'mean_sea_level']}
        prcVars = {'prate': ['PRATE', 'surface']}
        radVars = {'dlwrf': ['DLWRF', 'surface'],
                   'dswrf': ['DSWRF', 'surface']}

        lon, lat, idx_ymin, idx_ymax, idx_xmin, idx_xmax = self.modified_latlon() 
        path = pathlib.Path(outdir)
        path.mkdir(parents=True, exist_ok=True)

        self.cycle = start_date.hour
        base_url = f's3://hrrrzarr/sfc/{self.start_date.strftime("%Y%m%d")}/' \
            + f'{self.start_date.strftime("%Y%m%d")}_{self.cycle:02d}z_fcst.zarr'

        if air:
            xdat = []
            xdat.append(self.lon_DataArray(lon1D=lon))
            xdat.append(self.lat_DataArray(lat1D=lat))
            for key, value in airVars.items():
                url = f'{base_url}/{value[1]}/{value[0]}/{value[1]}/'
                logger.info(url)
                ds = xr.open_zarr(fsspec.get_mapper(url, anon=True), consolidated=False)
                ds = ds.rename({'projection_y_coordinate': 'ny_grid'})
                ds = ds.rename({'projection_x_coordinate': 'nx_grid'})
                val = ds[value[0]][:, idx_ymin:idx_ymax+1, idx_xmin:idx_xmax+1].astype('float32')
                xdat.append(xr.DataArray(val, dims=("time", "ny_grid", "nx_grid",), name=key))

                #add variable time
                times = ds['time'].values.astype('float32')
               
                #zarr data already discard t00z, which has all junk values
                times = (times+1)/24
                xdat.append(self.time_DataArray(time1D=times))

            fout = xr.merge(xdat)
            fout = fout.rename_vars({'time2': 'time'})
            #fout.to_netcdf(path / f'hrrr_{start_date.strftime("%Y%m%d")}{cycle:02d}.nc', 'w', unlimited_dims='time')
            fout.to_netcdf(path / f'sflux_air_2.{stack:04d}.nc', 'w', unlimited_dims='time')
            fout.close()

        if prc:
            xdat = []
            xdat.append(self.lon_DataArray(lon1D=lon))
            xdat.append(self.lat_DataArray(lat1D=lat))
            for key, value in prcVars.items():
                url = f'{base_url}/{value[1]}/{value[0]}/{value[1]}/'
                logger.info(url)
                ds = xr.open_zarr(fsspec.get_mapper(url, anon=True), consolidated=False)
                ds = ds.rename({'projection_y_coordinate': 'ny_grid'})
                ds = ds.rename({'projection_x_coordinate': 'nx_grid'})
                val = ds[value[0]][:, idx_ymin:idx_ymax+1, idx_xmin:idx_xmax+1].astype('float32')
                xdat.append(xr.DataArray(val, dims=("time", "ny_grid", "nx_grid",), name=key))

                #add variable time
                times = ds['time'].values.astype('float32')
                times = (times+1)/24
                #bdate = pd.to_datetime(start_date + timedelta(hours=cycle)).strftime('%Y %m %d %H').split(' ')
                #bdate = [int(q) for q in bdate[:4]] + [0]
                xdat.append(self.time_DataArray(time1D=times))
            fout = xr.merge(xdat)
            fout = fout.rename_vars({'time2': 'time'})
            #fout.to_netcdf(path / f'hrrr_{start_date.strftime("%Y%m%d")}{cycle:02d}.nc', 'w', unlimited_dims='time')
            fout.to_netcdf(path / f'sflux_prc_2.{stack:04d}.nc', 'w', unlimited_dims='time')
            fout.close()

        if rad:
            xdat = []
            xdat.append(self.lon_DataArray(lon1D=lon))
            xdat.append(self.lat_DataArray(lat1D=lat))
            for key, value in radVars.items():
                url = f'{base_url}/{value[1]}/{value[0]}/{value[1]}/'
                logger.info(url)
                ds = xr.open_zarr(fsspec.get_mapper(url, anon=True), consolidated=False)
                ds = ds.rename({'projection_y_coordinate': 'ny_grid'})
                ds = ds.rename({'projection_x_coordinate': 'nx_grid'})
                val = ds[value[0]][:, idx_ymin:idx_ymax+1, idx_xmin:idx_xmax+1].astype('float32')
                #DSWRF are all junk values in this dataset
                #if np.sum(np.isnan(val)) > 0:
                #    val[:] = 0.0
                xdat.append(xr.DataArray(val, dims=("time", "ny_grid", "nx_grid",), name=key))

                #add variable time
                times = ds['time'].values.astype('float32')
                times = (times+1)/24
                #bdate = pd.to_datetime(start_date + timedelta(hours=cycle)).strftime('%Y %m %d %H').split(' ')
                #bdate = [int(q) for q in bdate[:4]] + [0]
                xdat.append(self.time_DataArray(time1D=times))
            fout = xr.merge(xdat)
            fout = fout.rename_vars({'time2': 'time'})
            #fout.to_netcdf(path / f'hrrr_{start_date.strftime("%Y%m%d")}{cycle:02d}.nc', 'w', unlimited_dims='time')
            fout.to_netcdf(path / f'sflux_rad_2.{stack:04d}.nc', 'w', unlimited_dims='time')
            fout.close()
        
        #fout = xr.merge(xdat)
        #fout=fout.rename_vars({'time2': 'time'})
        #fout.to_netcdf(path / f'hrrr_{start_date.strftime("%Y%m%d")}{cycle:02d}.nc', 'w', unlimited_dims='time')

    def write(
        self,
        outdir,
        start_date: datetime = None,
        rnday: Union[float, timedelta] = 2,
        air: bool = True,
        prc: bool = False,
        rad: bool = False,
    ):

        start_date = nearest_cycle() if start_date is None else start_date

        if not isinstance(rnday, timedelta):
            try:
                rnday = timedelta(days=rnday)
            except:
                raise TypeError("rnday must be either float or datetime.timedelta!")

        timevector = np.arange(
            start_date,
            rnday,
            np.timedelta64(1, "D"),
            dtype="datetime64",
        ).astype(datetime)

        for i, date in enumerate(timevector):
            self.gen_sflux(outdir, date, i+1, air=air, prc=prc, rad=rad)

    def modified_latlon(self):
        ds = xr.open_dataset(self.grid_file)
        lon = ds['longitude'].astype("float32")
        lat = ds['latitude'].astype("float32")
        xmin = self.bbox.xmin
        xmax = self.bbox.xmax
        ymin = self.bbox.ymin
        ymax = self.bbox.ymax
        lon_idxs=(lon.values >= xmin) & (lon.values <= xmax)
        lat_idxs=(lat.values >= ymin) & (lat.values <= ymax+2.0)
        idxs = lon_idxs & lat_idxs
        idxs = np.argwhere(idxs)
        idx_ymin = np.min(idxs[:,0])
        idx_ymax = np.max(idxs[:,0])
        idx_xmin = np.min(idxs[:,1])
        idx_xmax = np.max(idxs[:,1])
        #print(f'idx_ymin is {idx_ymin}, idx_ymax is {idx_ymax}, idx_xmin is {idx_xmin}, idx_xmax is {idx_xmax}')
        return lon[idx_ymin:idx_ymax+1, idx_xmin:idx_xmax+1], lat[idx_ymin:idx_ymax+1, idx_xmin:idx_xmax+1], idx_ymin, idx_ymax, idx_xmin, idx_xmax


    def lon_DataArray(self, lon1D):
        return xr.DataArray(
            data=lon1D, 
            dims=("ny_grid", "nx_grid",), 
            name="lon",
            attrs=dict(
                long_name = "Longitude",
                standard_name = "longitude",
                units = "degrees_east"
            )
        )

    def lat_DataArray(self, lat1D):
        return xr.DataArray(
            data=lat1D, 
            dims=("ny_grid", "nx_grid",), 
            name="lat",
            attrs=dict(
                long_name = "Latitude",
                standard_name = "latitude",
                units = "degrees_north"
            )
        )

    def time_DataArray(self, time1D):
        #bdate = pd.to_datetime(self.start_date + timedelta(hours=self.cycle)).strftime('%Y %m %d %H').split(' ')
        bdate = pd.to_datetime(self.start_date).strftime('%Y %m %d %H').split(' ')
        bdate = [int(q) for q in bdate[:4]] + [0]
        
        return xr.DataArray(
            time1D, 
            dims=('time'), 
            name='time2',
            attrs=dict(
                long_name="Time",
                standard_name = "time",
                base_date = bdate,
                units = f"days since {self.start_date.year}-{self.start_date.month}-{self.start_date.day} {self.cycle:02d}:00 UTC"
            )
        )

    @property
    def s3(self):
        try:
            return self._s3
        except AttributeError:
            self._s3 = boto3.client(
                "s3", config = Config(signature_version=UNSIGNED))
            return self._s3

    @property
    def bucket(self):
        return "hrrrzarr"

    @property
    def grid_file(self):
        #if not hasattr(self, "_grid_file"):
        self._grid_file = "HRRR_latlon.h5"
        with open(self._grid_file, "wb") as f:
            self.s3.download_fileobj(f"{self.bucket}", "grid/HRRR_latlon.h5", f)
        return self._grid_file



from pathlib import Path
from argparse import Namespace

from shapely.geometry import Polygon, MultiPolygon
import geopandas as gpd

from pyschism.mesh.fgrid import ManningsN
from pyschism.mesh import Hgrid


class ManningsNCli:

    def __init__(self, args: Namespace):

        if args.sub_action == 'generate':

            outdir = args.out_dir
            if not outdir:
                outdir = Path(args.hgrid).parent
            outdir = Path(outdir)

            hgrid = Hgrid.open(args.hgrid, crs=args.hgrid_crs)

            if args.constant is not None:
                mann_obj = ManningsN.constant(hgrid, args.constant)

            elif args.linear is not None:
                # NOTE: args.linear can be empty list in which case
                # we need to use defaults
                keys = ["min_value", "max_value", "min_depth", "max_depth"]
                linear_args = {
                    keys[i]: val for i, val in enumerate(args.linear)}

                mann_obj = ManningsN.linear_with_depth(
                    hgrid, **linear_args)

            for i, (shpfile, value) in enumerate(args.region_specific):
                try:
                    shpfile = Path(shpfile)
                except TypeError:
                    raise TypeError(
                        f"Invalid input type ({type(value)}) for"
                        f" region shapefile path")

                try:
                    value = float(value)
                except ValueError:
                    raise TypeError(
                        f"Invalid input type ({type(value)}) for"
                        f" constant value across region")

                gdf = gpd.read_file(shpfile)

                if gdf.crs and mann_obj.crs and not gdf.crs.equals(mann_obj.crs):
                    gdf = gdf.to_crs(mann_obj.crs)
                multipolygon = gdf.unary_union
                if not isinstance(multipolygon, (Polygon, MultiPolygon)):
                    raise ValueError(
                        f"Invalid input shape for region {i}")

                mann_obj.add_region(multipolygon, value)

            mann_obj.write(outdir/'manning.gr3', overwrite=args.overwrite)

            return

        raise NotImplementedError(f'Unhandled CLI action: {args.action}.')


def manning_subparser(subparsers):

    # creating manning
    manning = subparsers.add_parser('manning')

    sub_actions = manning.add_subparsers(dest='sub_action')

    gen_mann = sub_actions.add_parser('generate')

    gen_mann.add_argument('--hgrid', required=True)
    gen_mann.add_argument('--hgrid-crs')
    gen_mann.add_argument('--out-dir', required=False)
    gen_mann.add_argument('--overwrite', action='store_true')

    gen_mode_grp = gen_mann.add_mutually_exclusive_group(required=True)
    gen_mode_grp.add_argument('--constant', type=float)
    gen_mode_grp.add_argument('--linear', nargs='*', type=float)

    gen_mann.add_argument(
        '--region-specific', nargs=2, action='append',
        default=list())


add_manning = manning_subparser



#!/usr/bin/env python
from datetime import datetime
import logging
import pathlib

from pyschism.forcing.source_sink.nwm import NationalWaterModel, NWMElementPairings
from pyschism.mesh import Hgrid

logging.basicConfig(
    format="[%(asctime)s] %(name)s %(levelname)s: %(message)s",
    force=True,
)

logging.getLogger("pyschism").setLevel(logging.DEBUG)

output_directory = pathlib.Path("outputs")

startdate = datetime(2018, 8, 24)

rnday = 36

hgrid = Hgrid.open('hgrid.ll')

sources_pairings = pathlib.Path("static/sources.json")
sinks_pairings = pathlib.Path("static/sinks.json")

if all([sources_pairings.is_file(), sinks_pairings.is_file()]) is False:
    pairings = NWMElementPairings(hgrid)
    sources_pairings.parent.mkdir(exist_ok=True, parents=True)
    pairings.save_json(sources=sources_pairings, sinks=sinks_pairings)

else:
    pairings = NWMElementPairings.load_json(
        hgrid, sources_pairings, sinks_pairings)

nwm = NationalWaterModel(
    # aggregation_radius=4000,
    pairings=pairings,
    cache='NWM_v2.0'
)
start = datetime.now()
nwm.write(
    output_directory,
    hgrid,
    startdate,
    rnday,
    overwrite=True,
)
print(f'Write out took {datetime.now() - start}')



from argparse import Namespace
import pathlib

from pyschism.mesh.parsers import sms2dm
from pyschism.mesh import Gr3


class Sms2grdCli:

    def __init__(self, args: Namespace):

        if args.output_path.exists() and args.overwrite is not True:
            raise IOError('File exists and overwrite is not True.')

        Gr3(**sms2dm_to_gr3(sms2dm.read(args.sms2dm))).write(
            args.output_path, args.overwrite)

    @staticmethod
    def add_subparser_action(subparsers):
        add_sms2grd(subparsers.add_parser('sms2grd'))


def sms2dm_to_gr3(sms2dm):
    nodes = {id: (coords, -value) for id, (coords, value)
             in sms2dm['ND'].items()}
    elements = {}
    if 'E3T' in sms2dm:
        elements.update({i+1: geom for i, geom
                         in enumerate(sms2dm['E3T'].values())})

    if 'E4Q' in sms2dm:
        offset = len(elements)
        elements.update({i+1+offset: geom for i, geom
                         in enumerate(sms2dm['Q4T'].values())})
    return {
        'nodes': nodes,
        'elements': elements,
        # 'boundaries': boundaries,
        'description': '',
        'crs': 'epsg:4326',
    }


def add_sms2grd(parser):
    parser.add_argument('sms2dm', type=pathlib.Path)
    parser.add_argument('output_path', type=pathlib.Path)
    parser.add_argument('--overwrite', action='store_true')



import pathlib
import re
import tempfile
import urllib.request

import f90nml


def typecast_value(value: str):
    if "real(" in value:
        value, exponent = value.split("(")[-1].split(",")[0].lower().split("d")
        return float(f"{value}e{exponent}")
    elif "_rkind" in value:
        return float(value.split("_rkind")[0])
    elif "'" in value or '"' in value:
        return value[1:-1]
    elif re.match("^[0-9].[0-9]*[de][+-]*[0-9]*", value):
        value, exponent = value.split("d")
        return float(f"{value}e{exponent}")
    elif "pi" in value:
        return value
    elif re.match("[+-]?([0-9]*[.])?[0-9]+", value):
        if "." in value:
            return float(value)
        else:
            return int(value)
    else:
        return value


class NamelistParser:
    def __new__(cls, schism_init_f90: str, component):
        schism_init_f90 = schism_init_f90.split("\n")
        for i, line in enumerate(schism_init_f90):
            line = line.lower()
            if f"namelist/{component}/" in line.replace(" ", ""):
                declared_variables = (
                    line.strip()
                    .split("namelist")[1]
                    .strip()
                    .split(f"/{component}/")[1]
                    .strip()
                    .split("&")[0]
                    .strip()
                    .split(",")[:-1]
                )
                break
        line = "&&"

        while line[0] == line[-1] == "&":
            line = schism_init_f90[i + 1].strip()
            declared_variables.extend(
                [var for var in line.replace("&", "").strip().split(",") if var]
            )
            i += 1
        defaults = {var: None for var in declared_variables}
        for var in declared_variables:
            for line in schism_init_f90[i + 1 :]:
                if "!" in line:
                    continue
                for source_var in line.split(";"):
                    source_var = source_var.strip()
                    if (
                        re.match(
                            rf"^{var}[ ]*=[ ]*[-_A-Za-z0-9 .\(\),]*[;]*", source_var
                        )
                        is not None
                    ):
                        value = (
                            [x for x in line.replace(" ", "").split(";") if var in x]
                            .pop()
                            .split("=")[-1]
                        )
                        defaults.update({var: typecast_value(value)})
                        break

        return defaults


class CoreParser(NamelistParser):
    def __new__(cls, schism_init_f90):
        return super().__new__(cls, schism_init_f90, "core")


class OptParser(NamelistParser):
    def __new__(cls, schism_init_f90):
        return super().__new__(cls, schism_init_f90, "opt")


class SchoutParser(NamelistParser):
    def __new__(cls, schism_init_f90):
        return super().__new__(cls, schism_init_f90, "schout")


class SchismInitParser:
    def __init__(self, branch="master"):
        url = f"https://raw.githubusercontent.com/schism-dev/schism/{branch}/src/Hydro/schism_init.F90"
        response = urllib.request.urlopen(url)
        self.schism_init_f90 = response.read().decode("utf-8")

    @property
    def opt(self):
        return OptParser(self.schism_init_f90)

    @property
    def core(self):
        return CoreParser(self.schism_init_f90)

    @property
    def schout(self):
        return SchoutParser(self.schism_init_f90)


def read_schism_param_sample_patched(schism_param_file):
    schism_param_sample = pathlib.Path(schism_param_file).read_text().decode('utf-8').split('\n')
    pattern = r"^[! _A-Za-z0-9]+\([*\d]*(\d+)[^\d]*\)[ ]*="
    var_collection = []
    for i, line in enumerate(schism_param_sample):
        if re.match(pattern, line):
            if line[0] != "!":
                schism_param_sample[i] = line = "!" + line
            varname = line.strip().split("=")[0].strip()[1:].strip().split("(")[0]
            var_collection.append(varname)
    declared_vars = list(set(var_collection))
    for var in declared_vars:
        varcount = var_collection.count(var)
        for i, line in enumerate(list(schism_param_sample)):
            if f"{var}({varcount})" in line:
                if varcount > 1:
                    schism_param_sample.insert(
                        i + 1, f"{var}(1:{varcount}) = " + ", ".join(varcount * ["0"])
                    )
                else:
                    schism_param_sample.insert(i + 1, f"{var}({varcount}) = 0")
                break
    return '\n'.join(schism_param_sample)

class Singleton(type):
    def __init__(cls, name, bases, dict):
        super(Singleton, cls).__init__(name, bases, dict)
        cls.instance = None

    def __call__(cls, *args, **kw):
        if cls.instance is None:
            cls.instance = super(Singleton, cls).__call__(*args, **kw)
        return cls.instance


class GitParamTemplate(metaclass=Singleton):
    def __init__(self, branch="master"):
        url = f"https://raw.githubusercontent.com/schism-dev/schism/{branch}/sample_inputs/param.nml"
        response = urllib.request.urlopen(url)
        tmpfile = tempfile.NamedTemporaryFile()
        with open(tmpfile.name) as fh:
            fh.write(response.read().decode("utf-8"))
        # schism_param_sample = response.read().decode("utf-8").split("\n")
        schism_param_sample = read_schism_param_sample_patched(tmpfile.name)
        tmpfile1 = tempfile.NamedTemporaryFile()
        with open(tmpfile1.name, "w") as f:
            f.write("\n".join(schism_param_sample))
        base_param = f90nml.read(tmpfile1.name)

        def update_param_component(name, component):
            for source_code_var, source_code_value in component.items():
                if source_code_var in base_param[name]:
                    if isinstance(base_param[name][source_code_var], list):
                        continue
                    if source_code_value != base_param[name][source_code_var]:
                        base_param[name][source_code_var] = source_code_value

        schism_init_f90 = SchismInitParser()
        update_param_component("core", schism_init_f90.core)
        update_param_component("opt", schism_init_f90.opt)
        update_param_component("schout", schism_init_f90.schout)
        tmpfile2 = tempfile.NamedTemporaryFile()
        f90nml.patch(tmpfile1.name, base_param, tmpfile2.name)
        with open(tmpfile2.name) as f:
            self.schism_param_sample = "".join(f.readlines())
        self.tmpfile = tmpfile2
        self.schism_init_f90 = schism_init_f90

    def __str__(self):
        return self.schism_param_sample

    def write(self, path):
        pathlib.Path(path).touch()
        f90nml.patch(self.path, str(self), path)

    @property
    def core(self):
        return self.schism_init_f90.core

    @property
    def opt(self):
        return self.schism_init_f90.opt

    @property
    def schout(self):
        return self.schism_init_f90.schout

    @property
    def path(self):
        return pathlib.Path(self.tmpfile.name)



from datetime import datetime, timedelta
from enum import Enum
import os
import pathlib
import subprocess
import tempfile
from typing import List, Union  # , Iterable

# import numpy as np

from pyschism import dates
from pyschism.enums import Stratification
from pyschism.hotstart import Hotstart

# from pyschism.forcing import Tides, Hydrology
from pyschism.forcing.source_sink import SourceSink
from pyschism.forcing.nws.base import NWS
from pyschism.forcing.nws.nws2 import NWS2
from pyschism.forcing.nws.best_track import BestTrackForcing

# from pyschism.forcing.baroclinic import BaroclinicForcing
from pyschism.forcing.bctides import Bctides
from pyschism.makefile import MakefileDriver
from pyschism.mesh import Hgrid, Vgrid, Fgrid, gridgr3, prop
from pyschism.mesh.fgrid import ManningsN, DragCoefficient
from pyschism.param import Param
from pyschism.server.base import ServerConfig
from pyschism.stations import Stations


def raise_type_error(argname, obj, cls):
    raise TypeError(
        f"Argument {argname} must be of type {cls}, not " f"type {type(obj)}."
    )


class ModelForcings:
    def __init__(self, bctides=None, nws=None, source_sink=None, waves=None):
        self.bctides = bctides
        self.nws = nws
        self.source_sink = source_sink
        self.waves = waves

    def write(
        self,
        driver: "ModelDriver",
        output_directory,
        overwrite,
    ):

        if self.bctides is not None:
            self.bctides.write(
                output_directory,
                start_date=driver.param.opt.start_date,
                rnday=driver.param.core.rnday,
                overwrite=True,
            )

        if self.nws is not None:
            if isinstance(self.nws, NWS2):
                self.nws.write(
                    output_directory,
                    start_date=driver.param.opt.start_date,
                    end_date=driver.param.core.rnday,
                    overwrite=overwrite,
                    bbox=driver.config.hgrid.get_bbox(output_type="bbox"),
                    prc=True if driver.config.vgrid.is3D() is True else False,
                    rad=True if driver.config.vgrid.is3D() is True else False,
                )
            elif isinstance(self.nws, BestTrackForcing):
                self.nws.write(
                    output_directory,
                    overwrite=overwrite,
                )

        if self.source_sink is not None:
            self.source_sink.write(
                output_directory,
                driver.config.hgrid,
                start_date=driver.param.opt.start_date,
                end_date=driver.param.core.rnday,
                overwrite=overwrite,
            )

        if self.waves is not None:
            self.waves.write()

class Gr3FieldTypes(Enum):
    ALBEDO = gridgr3.Albedo
    DIFFMIN = gridgr3.Diffmin
    DIFFMAX = gridgr3.Diffmax
    WATERTYPE = gridgr3.Watertype
    WINDROT = gridgr3.Windrot
    ESTUARY = gridgr3.Estuary

    @classmethod
    def _missing_(cls, name):
        raise ValueError(f"{name} is not a valid {gridgr3.Gr3Field} type.")


class PropTypes(Enum):
    FLUXFLAG = prop.Fluxflag
    TVDFLAG = prop.Tvdflag


class ModelDriver:
    def __init__(
        self,
        config: "ModelConfig",
        dt: Union[float, timedelta],
        rnday: Union[float, timedelta],
        start_date: datetime = None,
        dramp: Union[float, timedelta] = None,
        drampbc: Union[float, timedelta] = None,
        dramp_ss: Union[float, timedelta] = None,
        drampwafo: Union[float, timedelta] = None,
        drampwind: Union[float, timedelta] = None,
        elev_ic=None,
        temp_ic=None,
        salt_ic=None,
        nspool: Union[int, timedelta] = None,
        ihfskip: int = None,
        nhot_write: Union[int, timedelta] = None,
        stations: Stations = None,
        hotstart: Union[Hotstart, "ModelDriver"] = None,
        server_config: ServerConfig = None,
        param_template=None,
        **surface_outputs,
    ):
        self.config = config
        self.param = Param()

        # set core parameters
        self.param.core.dt = dt
        self.param.core.rnday = rnday
        self.param.core.nspool = nspool if nspool is not None else self.param.core.rnday
        self.param.core.ihfskip = (
            ihfskip if ihfskip is not None else timedelta(days=self.param.core.rnday)
        )
        # if self.config.vgrid.is2D():
        #     self.param.core.ibc = Stratification.BAROTROPIC
        #     if self.config.forcings.baroclinic is not None:
        #         self.param.core.ibtp = 1
        # else:
        #     self.param.core.ibc = Stratification.BAROCLINIC
        self.param.core.ibc = self.config.stratification

        # TODO: must also set msc2/mdc2 here.

        # set opt
        self.param.opt.start_date = start_date
        self.param.opt.ics = 2 if self.config.hgrid.crs.is_geographic is True else 1
        self.param.opt.ncor = 1 if self.param.opt.ics == 2 else 0
        self.param.opt.dramp = dramp
        self.param.opt.drampbc = drampbc
        self.param.opt.dramp_ss = dramp_ss
        self.param.opt.drampwafo = drampwafo
        self.param.opt.drampwind = drampwind

        # set friction parameters
        self.param.opt.nchi = self.config.fgrid.nchi
        if self.config.fgrid.nchi == -1:
            self.param.opt.hmin_man = self.config.fgrid.hmin_man

        elif self.config.fgrid.nchi == 1:
            self.param.opt.dbz_min = self.config.fgrid.dbz_min
            self.param.opt.dbz_decay = self.config.fgrid.dbz_decay

        self.param.schout.nhot_write = (
            self.param.core.ihfskip if nhot_write is None else nhot_write
        )

        self.tmpdir = tempfile.TemporaryDirectory()
        self.outdir = pathlib.Path(self.tmpdir.name)

        self.stations = stations

        self.server_config = server_config

        self.hotstart = hotstart

        self.elev_ic = elev_ic

        # set flag_ic
        # flag_ic(1) = 1 !T
        self.temp_ic = temp_ic
        # flag_ic(2) = 1 !S
        self.salt_ic = salt_ic

        # TODO: init tracers (flag_ic[2:])

        if self.config.forcings.nws is not None:
            if isinstance(self.config.forcings.nws, NWS2):
                if self.config.forcings.nws.windrot is None:
                    self.config.forcings.nws.windrot = gridgr3.Windrot.default(
                        self.config.hgrid
                    )
            elif isinstance(self.config.forcings.nws, BestTrackForcing):
                # Writing the rotation file is still needed, but it's
                # not meaningful for BestTrackForcing
                self.config.forcings.nws.windrot = gridgr3.Windrot.default(
                    self.config.hgrid
                )
                self.param.opt.model_type_pahm = self.config.forcings.nws.model.value

            self.param.opt.wtiminc = self.param.core.dt
            self.param.opt.nws = self.config.forcings.nws.dtype.value

        if self.config.forcings.source_sink is not None:
            if self.elev_ic is None:
                if self.hotstart is None:
                    self.elev_ic = True

        # TODO: init template
        self.param_template = param_template

        for var in self.param.schout.surface_output_vars:
            val = surface_outputs.pop(var) if var in surface_outputs else None
            if val is not None:
                setattr(self.param.schout, var, val)

        if len(surface_outputs) > 0:
            raise TypeError(
                "ModelDriver() got an unexpected keyword arguments"
                f" {list(surface_outputs)}."
            )

    def run(
        self,
        output_directory: Union[str, os.PathLike] = None,
        overwrite=False,
        use_param_template=False,
    ):
        self.outdir = (
            pathlib.Path(output_directory)
            if output_directory is not None
            else self.outdir
        )
        self.write(
            self.outdir, overwrite=overwrite, use_param_template=use_param_template
        )
        # Make sure we are using a fresh fatal.error file since we can't catch
        # blowup from mpiexec exit codes.
        error_file = self.outdir / "outputs/fatal.error"
        if error_file.exists() and overwrite is not True:
            raise IOError("File exists and overwrite is not True.")
        if error_file.exists() and overwrite is True:
            error_file.unlink()
        subprocess.check_call(["make", "run"], cwd=self.outdir)
        with open(error_file) as f:
            error = f.read()
        if "ABORT" in error:
            raise Exception(f"SCHISM exited with error:\n{error}")

    @property
    def config(self) -> "ModelConfig":
        return self._config

    @config.setter
    def config(self, config: "ModelConfig"):
        if not isinstance(config, ModelConfig):
            raise_type_error("config", config, ModelConfig)
        self._config = config

    @property
    def hotstart(self):
        return self._hotstart

    @hotstart.setter
    def hotstart(self, hotstart: Union[Hotstart, "ModelDriver", None]):

        if hotstart is None:
            pass

        else:
            if not isinstance(hotstart, Hotstart):
                if isinstance(hotstart, self.__class__):
                    hotstart = Hotstart.combine(hotstart.outdir / "outputs")
                else:
                    raise TypeError(
                        f"Argument hotstart must be of type {Hotstart}, "
                        f"{self.__class__} or None, not type {type(hotstart)}."
                    )

        if hotstart is not None:
            self.param.opt.ihot = 1

        self._hotstart = hotstart

    @property
    def stations(self) -> Stations:
        return self._stations

    @stations.setter
    def stations(self, stations: Union[Stations, None]):
        if stations is not None:
            if not isinstance(stations, Stations):
                raise_type_error("stations", stations, Stations)
        self._stations = stations

    def write(
        self,
        output_directory,
        overwrite=False,
        hgrid=True,
        vgrid=True,
        fgrid=True,
        param=True,
        bctides=True,
        nws=True,
        stations=True,
        use_param_template=False,
        albedo=True,
        diffmax=True,
        diffmin=True,
        watertype=True,
        windrot=True,
        shapiro=True,
        fluxflag=False,
        tvdflag=True,
        elev_ic=True,
        temp_ic=True,
        salt_ic=True,
        # rtofs=True,
        hydrology=True,
    ):
        """Writes to disk the full set of input files necessary to run SCHISM."""

        self.outdir = (
            pathlib.Path(output_directory)
            if output_directory is not None
            else self.outdir
        )

        if not (self.outdir / "outputs").exists():
            (self.outdir / "outputs").mkdir(parents=True, exist_ok=overwrite)

        if hgrid is not False:
            hgrid = "hgrid.gr3" if hgrid is True else hgrid
            self.config.hgrid.write(self.outdir / hgrid, overwrite)
            if self.param.opt.ics == 2:
                _original_dir = os.getcwd()
                os.chdir(self.outdir)  # pushd
                try:
                    os.remove("hgrid.ll")
                except OSError:
                    pass
                os.symlink(hgrid, "hgrid.ll")
                os.chdir(_original_dir)  # popd

        if vgrid is not False:
            vgrid = "vgrid.in" if vgrid is True else vgrid
            self.config.vgrid.write(self.outdir / vgrid, overwrite)

        if fgrid is not False:
            fgrid = f"{self.config.fgrid.fname}" if fgrid is True else fgrid
            self.config.fgrid.write(self.outdir / fgrid, overwrite)

        if param is not False:
            param = "param.nml" if param is True else param
            self.param.write(
                self.outdir / param, overwrite, use_template=use_param_template
            )

        def obj_write(var, obj, default_filename, overwrite):
            if var is not False and obj is not None:
                obj.write(
                    self.outdir / default_filename if var is True else var, overwrite
                )

        obj_write(albedo, self.config.albedo, "albedo.gr3", overwrite)
        obj_write(diffmin, self.config.diffmin, "diffmin.gr3", overwrite)
        obj_write(diffmax, self.config.diffmax, "diffmax.gr3", overwrite)
        obj_write(watertype, self.config.watertype, "watertype.gr3", overwrite)
        obj_write(windrot, self.config.windrot, "windrot_geo2proj.gr3", overwrite)
        obj_write(shapiro, self.config.shapiro, "shapiro.gr3", overwrite)
        obj_write(fluxflag, self.config.fluxflag, "fluxflag.prop", overwrite)
        obj_write(tvdflag, self.config.tvdflag, "tvd.prop", overwrite)
        obj_write(temp_ic, self.temp_ic, "temp.ic", overwrite)
        obj_write(salt_ic, self.salt_ic, "salt.ic", overwrite)
        obj_write(elev_ic, self.elev_ic, "elev.ic", overwrite)
        obj_write(stations, self.stations, "station.in", overwrite)

        self.config.forcings.write(
            self, output_directory, overwrite,
        )

        MakefileDriver(self.server_config, hotstart=self.hotstart).write(
            self.outdir / "Makefile", overwrite
        )

    @property
    def outdir(self):
        return self._outdir

    @outdir.setter
    def outdir(self, outdir):
        self.config.outdir = outdir
        self._outdir = outdir

    @property
    def elev_ic(self):
        return self._elev_ic

    @elev_ic.setter
    def elev_ic(self, elev_ic: Union[None, bool, gridgr3.ElevIc]):
        if elev_ic is not None:
            if not isinstance(elev_ic, (gridgr3.ElevIc, bool)):
                raise_type_error("elev_ic", elev_ic, gridgr3.ElevIc)
            if elev_ic is True:
                elev_ic = gridgr3.ElevIc.default(self.config.hgrid)
            self.param.opt.ic_elev = 1
            if elev_ic is False:
                self.param.opt.ic_elev = 0
        self._elev_ic = elev_ic

    @elev_ic.deleter
    def elev_ic(self):
        self.param.opt.ic_elev = None
        self._elev_ic = None


class Gridgr3Descriptor:
    def __init__(self, gridgr3_type: Gr3FieldTypes):
        self.name = gridgr3_type.name
        self.type = gridgr3_type.value
        self.gr3field = None

    def __set__(self, obj, val: Union[gridgr3.Gr3Field, None]):
        if not isinstance(val, self.type) and val is not None:
            raise ValueError(
                f"Argument {self.name.lower()} must be of type {self.type} "
                f"not type {type(val)}."
            )
        self.gr3field = val

    def __get__(self, obj, val):
        return self.gr3field


class PropTypeDescriptor:
    def __init__(self, prop_type: PropTypes):
        self.name = prop_type.name
        self.type = prop_type.value
        self.prop = None

    def __set__(self, obj, val: Union[prop.Prop, None]):
        if not isinstance(val, self.type) and val is not None:
            raise ValueError(
                f"Argument {self.name.lower()} must be of type {self.type} "
                f"not type {type(val)}."
            )
        self.prop = val

    def __get__(self, obj, val):
        return self.prop


class ModelConfigMeta(type):
    def __new__(meta, name, bases, attrs):
        # attrs['forcings'] = ModelForcings()
        attrs["start_date"] = dates.StartDate()
        attrs["end_date"] = dates.EndDate()
        attrs["spinup_time"] = dates.SpinupTime()
        for gr3field_type in Gr3FieldTypes:
            name = gr3field_type.name.lower()
            attrs[name] = Gridgr3Descriptor(gr3field_type)
        for prop_type in PropTypes:
            name = prop_type.name.lower()
            attrs[name] = PropTypeDescriptor(prop_type)
        return type(name, bases, attrs)


class ModelConfig(metaclass=ModelConfigMeta):
    """Class representing a SCHISM model configuration.

    This class combines the horizontal grid (hgrid), vertical grid (vgrid)
    and friction/drag grids (fgrid). Additionally, this class holds
    information about forcings.

    Args:
        hgrid: :class:`pyschism.mesh.Hgrid` instance.
        vgrid: :class:`pyschism.mesh.Vgrid` instance.
        fgrid: :class:`pyschism.mesh.Fgrid` derived instance.
    """

    def __init__(
        self,
        hgrid: Hgrid,
        vgrid: Vgrid = None,
        fgrid: Fgrid = None,
        flags = None,
        constituents = 'major',
        database = 'tpxo',
        add_earth_tidal = True,
        ethconst = None,
        vthconst = None,
        tthconst = None,
        sthconst = None,
        tobc = None,
        sobc = None,
        relax = None, 
        # itrtype: itrtype.Itrtype = None,
        nws: NWS = None,
        source_sink: Union[List[SourceSink], SourceSink] = None,
        waves=None,
        stratification: Union[int, str, Stratification] = None,
        albedo: gridgr3.Albedo = None,
        diffmin: gridgr3.Diffmin = None,
        diffmax: gridgr3.Diffmax = None,
        watertype: gridgr3.Watertype = None,
        estuary: gridgr3.Estuary = None,
        shapiro: gridgr3.Shapiro = None,
        fluxflag: prop.Fluxflag = None,
        tvdflag: prop.Tvdflag = None,
    ):
        self.hgrid = hgrid
        self.vgrid = vgrid
        self.fgrid = fgrid
        self.flags = flags
        self.constituents = constituents
        self.add_earth_tidal = add_earth_tidal
        self.ethconst = ethconst
        self.vthconst = vthconst
        self.tthconst = tthconst
        self.sthconst = sthconst
        self.tobc = tobc
        self.sobc = sobc
        self.relax = relax
        # self.itrtype = itrtype
        self.nws = nws
        self.source_sink = source_sink
        self.waves = waves
        self.stratification = stratification
        self.albedo = albedo
        self.diffmin = diffmin
        self.diffmax = diffmax
        self.watertype = watertype
        self.shapiro = shapiro
        self.fluxflag = fluxflag
        self.tvdflag = tvdflag
        self.estuary = estuary

    def coldstart(
        self,
        timestep: Union[float, timedelta] = 150.0,
        start_date: datetime = None,
        end_date: Union[datetime, timedelta] = None,
        dramp: Union[float, timedelta] = None,
        drampbc: Union[float, timedelta] = None,
        dramp_ss: Union[float, timedelta] = None,
        drampwafo: Union[float, timedelta] = None,
        drampwind: Union[float, timedelta] = None,
        elev_ic: gridgr3.ElevIc = None,
        temp_ic: gridgr3.TempIc = None,
        salt_ic: gridgr3.TempIc = None,
        nspool: Union[int, timedelta] = None,
        ihfskip: int = None,
        nhot_write: Union[int, timedelta] = None,
        stations: Stations = None,
        server_config: ServerConfig = None,
        param_template=None,
        **surface_outputs,
    ) -> ModelDriver:

        if start_date is None:
            start_date = dates.nearest_cycle()

        if not isinstance(start_date, datetime):
            raise TypeError(
                f"Argument start_date must be of type {datetime} "
                f"or None, not type {type(start_date)}."
            )

        if end_date is None:
            end_date = self.forcings.maximum_end_date()

        if isinstance(end_date, timedelta):
            end_date = start_date + end_date

        if isinstance(end_date, (int, float)):
            end_date = start_date + timedelta(days=float(end_date))

        if not isinstance(end_date, datetime):
            raise TypeError(
                f"Argument end_date must be of type {datetime}, {timedelta}, "
                f"or None, not type {type(end_date)}."
            )

        return ModelDriver(
            self,
            dt=timestep,
            start_date=start_date,
            rnday=end_date - start_date,
            dramp=dramp,
            drampbc=drampbc,
            dramp_ss=dramp_ss,
            drampwafo=drampwafo,
            drampwind=drampwind,
            elev_ic=elev_ic,
            temp_ic=temp_ic,
            salt_ic=salt_ic,
            stations=stations,
            nspool=nspool,
            nhot_write=nhot_write,
            server_config=server_config,
            ihfskip=ihfskip,
            param_template=param_template,
            **surface_outputs,
        )

    def hotstart(
        self,
        hotstart: Union[Hotstart, ModelDriver],
        timestep: Union[float, timedelta] = 150.0,
        end_date: Union[datetime, timedelta] = None,
        nspool: Union[int, timedelta] = None,
        ihfskip: int = None,
        nhot_write: Union[int, timedelta] = None,
        stations: Stations = None,
        server_config: ServerConfig = None,
        **surface_outputs,
    ) -> ModelDriver:

        if isinstance(hotstart, ModelDriver):
            hotstart = Hotstart.combine(pathlib.Path(hotstart.outdir) / "outputs")

        if not isinstance(hotstart, Hotstart):
            raise TypeError(
                f"Argument hotstart must be of type {Hotstart}, "
                f"not type {type(hotstart)}."
            )

        if end_date is None:
            end_date = self.forcings.max_end_date()
            if end_date is None:
                raise ValueError("end_date is unbounded, must pass end_date argument.")
        if not isinstance(end_date, datetime):
            if isinstance(end_date, timedelta):
                end_date = hotstart.time + end_date
            else:
                end_date = hotstart.time + timedelta(days=float(end_date))

        return ModelDriver(
            self,
            dt=timestep,
            start_date=hotstart.time,
            rnday=end_date - hotstart.time,
            hotstart=hotstart,
            nspool=nspool,
            ihfskip=ihfskip,
            stations=stations,
            nhot_write=nhot_write,
            server_config=server_config,
            elev_ic=None,
            temp_ic=None,
            salt_ic=None,
            **surface_outputs,
        )

    @property
    def hgrid(self):
        return self._hgrid

    @hgrid.setter
    def hgrid(self, hgrid: Hgrid):
        if not isinstance(hgrid, Hgrid):
            raise_type_error("hgrid", hgrid, Hgrid)
        self._hgrid = hgrid

    @property
    def vgrid(self):
        return self._vgrid

    @vgrid.setter
    def vgrid(self, vgrid: Union[Vgrid, None]):
        if vgrid is None:
            vgrid = Vgrid.default()
        if not isinstance(vgrid, Vgrid):
            raise_type_error("vgrid", vgrid, Vgrid)
        self._vgrid = vgrid

    @property
    def fgrid(self):
        return self._fgrid

    @fgrid.setter
    def fgrid(self, fgrid: Union[Fgrid, None]):
        if fgrid is None:
            if self.vgrid.is2D():
                fgrid = ManningsN.linear_with_depth(self.hgrid)
            else:
                fgrid = DragCoefficient.linear_with_depth(self.hgrid)

        if not isinstance(fgrid, Fgrid):
            raise_type_error("fgrid", fgrid, Fgrid)

        if self.vgrid.is2D() is True and not isinstance(fgrid, ManningsN):
            raise TypeError(f"2D model must use {ManningsN} but got {type(fgrid)}.")

        self._fgrid = fgrid

    @property
    def timestep(self):
        return self._timestep

    @timestep.setter
    def timestep(self, timestep: Union[float, timedelta, None]):
        if timestep is None:
            self._timestep = timedelta(seconds=150.0)
        elif not isinstance(timestep, timedelta):
            self._timestep = timedelta(seconds=float(timestep))
        return self._timestep

    @property
    def outdir(self):
        return self._outdir

    @outdir.setter
    def outdir(self, outdir):
        self._outdir = pathlib.Path(outdir)

    @property
    def stratification(self):
        return self._stratification

    @stratification.setter
    def stratification(self, stratification: Union[str, Stratification, int, None]):
        if not isinstance(stratification, (str, Stratification, type(None))):
            raise TypeError(
                f"Argument stratification must be a str, {Stratification}, int (0 or 1), or None, not "
                f"type {type(stratification)}."
            )
        if stratification is None:
            if self.vgrid.is2D() is True:
                stratification = Stratification.BAROTROPIC
            else:
                stratification = Stratification.BAROCLINIC
        elif isinstance(stratification, (str, int)):
            stratification = Stratification(stratification)
        self._stratification = stratification

    @property
    def forcings(self):
        if not hasattr(self, "_forcings"):
            self._forcings = ModelForcings(
                bctides=self.bctides,
                nws=self.nws,
                source_sink=self.source_sink,
                waves=self.waves,
            )

        return self._forcings

    @property
    def bctides(self):
        if not hasattr(self, "_bctides"):
            self._bctides = Bctides(
                hgrid = self.hgrid,
                flags = self.flags,
                constituents = self.constituents,
                add_earth_tidal = self.add_earth_tidal,
                ethconst = self.ethconst,
                vthconst = self.vthconst,
                tthconst = self.tthconst,
                sthconst = self.sthconst,
                tobc = self.tobc,
                sobc = self.sobc,
                relax = self.relax,
                # cutoff_depth=self.cutoff_depth,
            )
        return self._bctides



import tempfile
from datetime import timedelta, datetime
from pathlib import Path
import unittest

from stormevents.nhc import VortexTrack

from pyschism.mesh import Hgrid
from pyschism.driver import ModelConfig
from pyschism.forcing.nws import BestTrackForcing

class ModelConfigurationTestCase(unittest.TestCase):
    def test_paramwind_from_stormname(self):

        hgrid = Hgrid.open(
            'https://raw.githubusercontent.com/geomesh/test-data/main/NWM/hgrid.ll'
        )

        meteo = BestTrackForcing(storm='Florence2018')

        config = ModelConfig(
            hgrid=hgrid,
            vgrid=None,
            fgrid=None,
            flags=[[3, 3, 0, 0]],
            nws=meteo,
        )

        driver = config.coldstart(
            start_date=datetime(2018, 9, 8),
            end_date=datetime(2018, 9, 18),
            timestep=timedelta(seconds=150),
            nspool=24,
        )

        with tempfile.TemporaryDirectory() as dn:
            tmpdir = Path(dn)
            driver.write(tmpdir / 'paramwind', overwrite=True)

    def test_paramwind_from_file(self):
        hgrid = Hgrid.open(
            'https://raw.githubusercontent.com/geomesh/test-data/main/NWM/hgrid.ll'
        )

        track = VortexTrack.from_storm_name('Florence', 2018)


        with tempfile.TemporaryDirectory() as dn:
            tmpdir = Path(dn)
            track.to_file(tmpdir / 'track.dat')

            meteo = BestTrackForcing.from_nhc_bdeck(nhc_bdeck=tmpdir / 'track.dat')
            config = ModelConfig(
                hgrid=hgrid,
                vgrid=None,
                fgrid=None,
                flags=[[3, 3, 0, 0]],
                nws=meteo,
            )

            driver = config.coldstart(
                start_date=datetime(2018, 9, 8),
                end_date=datetime(2018, 9, 18),
                timestep=timedelta(seconds=150),
                nspool=24,
            )

            driver.write(tmpdir / 'paramwind', overwrite=True)

if __name__ == '__main__':
    unittest.main()



from abc import ABC
from datetime import datetime, timedelta
import os
import pathlib
from typing import Dict, Union


# import cf
from netCDF4 import Dataset
import numpy as np
import xarray

from pyschism.enums import (
    IofWetdryVariables,
    IofZcorVariables,
    IofHydroVariables,
    IofDvdVariables,
    IofWwmVariables,
    IofGenVariables,
    IofAgeVariables,
    IofSedVariables,
    IofEcoVariables,
    IofIcmVariables,
    IofCosVariables,
    IofFibVariables,
    IofSed2dVariables,
    IofMarshVariables,
    IofIceVariables,
    IofAnaVariables,
)
from pyschism.mesh.base import Gr3


def get_stack_id_by_datetime(dt: datetime, flattened_timevector, stacks) -> str:
    output_stack = None
    for stack_id, stack_data in stacks.items():
        if dt in stack_data['timevector']:
            output_stack = stack_id
    if output_stack is None:
        raise ValueError(
            'The are no output slices corresponding to requested datetime:'
            f' {dt}. Available datetimes are '
            f'{flattened_timevector}.')
    return output_stack


# def aggregate_by_datetime(shape, filenames, n_local_to_global, hgrid, name, dt, stacks, flattened_timevector):
#     values = np.full(shape, np.nan)
#     stack_id = get_stack_id_by_datetime(dt, flattened_timevector, stacks)
#     local_time_index = get_local_time_index(dt, stacks, flattened_timevector)
#     for file in filenames:
#         cpu_id = file.name.split('_')[1]
#         _n_local_to_global = n_local_to_global[cpu_id]
#         local_node_ids = list(_n_local_to_global.keys())
#         global_node_ids = list(
#             map(lambda x: _n_local_to_global[x], local_node_ids))
#         idxs = list(map(hgrid.nodes.get_index_by_id, global_node_ids))
#         if file.name.endswith(f'{stack_id}.nc'):
#             values[idxs] = Dataset(file)[name][local_time_index, :]
#     return values


class OutputVariableCombiner(ABC):

    def __init__(self, parent, name, start_step=0):
        self.parent = parent
        self.name = name
        self.current_step = start_step

    @property
    def flattened_timevector(self):
        return self.parent.flattened_timevector

    @property
    def shape(self):
        return self.ncvar.shape[1:]

    @property
    def filenames(self):
        return self.parent.filenames

    @property
    def hgrid(self):
        return self.parent.hgrid

    @property
    def n_local_to_global(self):
        return self.parent.n_local_to_global

    @property
    def s_local_to_global(self):
        return self.parent.s_local_to_global

    @property
    def e_local_to_global(self):
        return self.parent.e_local_to_global

    @property
    def stacks(self):
        return self.parent.stacks

    def animation(
            self,
            save=False,
            fps=3,
            start_frame=0,
            end_frame=-1,
            figsize=None,
            wireframe=False,
            cmap='jet',
            levels=256,
            show=False,
            xmin=None,
            xmax=None,
            ymin=None,
            ymax=None,
            vmin=None,
            vmax=None,
    ):
        from matplotlib.animation import FuncAnimation
        import matplotlib.pyplot as plt
        self.current_step = start_frame
        fig = plt.figure(figsize)
        ax = fig.add_subplot(111)

        plt.tight_layout(pad=2)

        triangulation = self.hgrid.triangulation
        xmin = np.min(self.hgrid.x) if xmin is None else xmin
        xmax = np.max(self.hgrid.x) if xmax is None else xmax
        ymin = np.min(self.hgrid.y) if ymin is None else ymin
        ymax = np.max(self.hgrid.y) if ymax is None else ymax
        vmin = np.min(self.values) if vmin is None else vmin
        vmax = np.max(self.values) if vmax is None else vmax

        triangulation.set_mask(self.hgrid.elements.get_triangulation_mask(self.parent.wetdry_elem.values))

        def animate(index):
            self.current_step = index
            _ax = fig.get_axes()
            ax.clear()
            if len(_ax) > 1:
                cax = _ax[1]
                cax.cla()
            else:
                cax = None

            triangulation.set_mask(self.hgrid.elements.get_triangulation_mask(self.parent.wetdry_elem.values))

            if wireframe:
                ax.triplot(triangulation, color='k', linewidth=0.7)

            ax.tricontourf(
                triangulation,
                self.values,
                cmap=cmap,
                levels=levels,
                vmin=vmin,
                vmax=vmax
                )

            ax.set_ylim(ymin, ymax, auto=True)
            ax.set_xlim(xmin, xmax, auto=True)

            ax.set_xlabel('Longitude (°E)')
            ax.set_ylabel('Latitude (°N)')

            ax.set_title(self.flattened_timevector[index].strftime('%b %d, %Y %H:%M'))
            m = plt.cm.ScalarMappable(cmap=cmap)
            m.set_array(self.values[index])
            m.set_clim(vmin, vmax)
            cbar = fig.colorbar(
                m, cax=cax, format='%.1f',
                boundaries=np.linspace(vmin, vmax, levels)
            )
            ax.axis('scaled')
            # cbar = fig.colorbar(_ax)
            # cbar.ax.set_ylabel(f'{variable} [{unit}]', rotation=90)

        end_frame = end_frame % len(self.flattened_timevector) \
            if end_frame < 0 else end_frame
        start_frame = start_frame % len(self.flattened_timevector) \
            if start_frame < 0 else start_frame
        frames = range(start_frame, end_frame)
        anim = FuncAnimation(
            fig,
            animate,
            frames,
            blit=False
            )

        if save not in [False, None]:
            anim.save(
                pathlib.Path(save),
                writer='imagemagick',
                fps=fps
            )

        if show:
            plt.show()

        return anim

    def aggregate(self, dt: Union[datetime, timedelta, int]):
        if isinstance(dt, datetime):
            return self.aggregate_by_datetime(dt)
        # elif isinstance(dt, timedelta):
        #     return self.aggregate_by_timedelta(dt)
        elif isinstance(dt, int):
            return self.aggregate_by_index(dt)
        else:
            raise TypeError('Argument dt must be datetime, timedelta or int '
                            f'not type {type(dt)}')

    # def aggregate_parallel(self, nprocs=16):
    #     from time import time
    #     from multiprocessing import Pool
    #     start = time()
    #     with Pool(processes=nprocs) as p:
    #         # shape, filenames, n_local_to_global, hgrid, name, dt, stacks, flattened_timevector
    #         res = p.starmap(
    #             aggregate_by_datetime,
    #             [
    #                 (self.shape, self.filenames, self.n_local_to_global,
    #                  self.hgrid, self.name, dt, self.stacks,
    #                  self.flattened_timevector)
    #                 for dt in self.flattened_timevector
    #             ]
    #         )
    #     p.join()
    #     print(f'parallel took {time()-start}')
    #     start = time()
    #     for dt in self.flattened_timevector:
    #         aggregate_by_datetime(
    #             self.shape, self.filenames, self.n_local_to_global, self.hgrid,
    #             self.name, dt, self.stacks,
    #             self.flattened_timevector)
    #     print(f'serial took {time()-start}')
    #     # breakpoint()

    def aggregate_by_index(self, index):
        return self.aggregate_by_datetime(self.flattened_timevector[index])

    def aggregate_by_datetime(self, dt):

        stack_id = self.get_stack_id_by_datetime(dt)
        filenames = [file for file in self.filenames if str(file).endswith(f'{stack_id}.nc')]
        ncvar = Dataset(filenames[0])[self.name]
        shape = []
        if ncvar.dimensions[1] == 'nSCHISM_hgrid_node':
            shape.append(self.hgrid.nodes.values.shape[0])
        elif ncvar.dimensions[1] == 'nSCHISM_hgrid_face':
            shape.append(self.hgrid.elements.index.shape[0])
        else:
            raise Exception(f'Unhandled dimension1 {self.ncvar.dimensions[1]}')

        if len(ncvar.dimensions[1:]) > 1:
            raise Exception(f'Unhandled number of dimensions {self.ncvar.dimensions}')
        
        values = np.full(tuple(shape), np.nan)

        for i, file in enumerate(filenames):
            ncvar = Dataset(filenames[i])[self.name]
            cpu_id = file.name.split('_')[1]
            if ncvar.dimensions[1] == 'nSCHISM_hgrid_node':
                n_local_to_global = self.n_local_to_global[cpu_id]
                local_node_ids = list(n_local_to_global.keys())
                global_node_ids = list(
                    map(lambda x: n_local_to_global[x], local_node_ids))
                idxs = list(map(self.hgrid.nodes.get_index_by_id, global_node_ids))
            elif ncvar.dimensions[1] == 'nSCHISM_hgrid_face':
                e_local_to_global = self.e_local_to_global[cpu_id]
                local_elem_ids = list(e_local_to_global.keys())
                global_elem_ids = list(
                    map(lambda x: e_local_to_global[x], local_elem_ids))
                idxs = list(map(self.hgrid.elements.get_index_by_id, global_elem_ids))
            local_time_index = self.get_local_time_index(dt)
            values[idxs] = ncvar[local_time_index, :]

        return values

    def get_stack_id_by_datetime(self, dt: datetime) -> str:
        output_stack = None
        for stack_id, stack_data in self.stacks.items():
            if dt in stack_data['timevector']:
                output_stack = stack_id
        if output_stack is None:
            raise ValueError(
                'The are no output slices corresponding to requested datetime:'
                f' {dt}. Available datetimes are '
                f'{self.flattened_timevector}.')
        return output_stack

    def get_local_time_index(self, dt: datetime) -> str:
        for stack_data in self.stacks.values():
            if dt in stack_data['timevector']:
                return stack_data['timevector'].index(dt)
        raise ValueError(
                'The are no output slices corresponding to requested datetime:'
                f' {dt}. Available datetimes are '
                f'{self.flattened_timevector}.')

    def __iter__(self):
        return self

    def __next__(self):
        self.current_step += 1
        return self.values

    @property
    def values(self):
        return self.aggregate(self.current_step)

    @property
    def current_step(self):
        return self._current_step

    @current_step.setter
    def current_step(self, current_step: int):
        if current_step is None:
            current_step = self.timesteps[0]
        self._current_step = current_step

    # @property
    # def ncvar(self):
    #     return self.stacks

    # @property
    # def shape(self):
    #     if not hasattr(self, '_shape'):
    #         shape = [self.hgrid.values.shape[0]]
    #         if self.rank is not None:
    #             shape.append(self.rank)
    #         if self.nvrt is not None:
    #             shape.append(self.nvrt)
    #         self._shape = tuple(shape)
    #     return self._shape


class OutputsCollector:

    surface_output_vars = [
        IofWetdryVariables,
        IofZcorVariables,
        IofHydroVariables,
        IofDvdVariables,
        IofWwmVariables,
        IofGenVariables,
        IofAgeVariables,
        IofSedVariables,
        IofEcoVariables,
        IofIcmVariables,
        IofCosVariables,
        IofFibVariables,
        IofSed2dVariables,
        IofMarshVariables,
        IofIceVariables,
        IofAnaVariables,
    ]

    def __init__(self, path: Union[str, os.PathLike]):
        self.path = pathlib.Path(path)

        if not self.path.exists():
            raise ValueError(f'Specified directory {self.path} does not exist!')

        self.filenames = sorted(
            self.path.glob(r'schout_[0-9][0-9][0-9][0-9]_*.nc'))
        # self.fields = cf.read(self.filenames)
        nodes = {}
        elements = {}
        for file in sorted(self.path.glob(r'local_to_global_[0-9][0-9][0-9][0-9]')):
            with open(file) as f:
                ns_global, ne_global, np_global, nvrt, nproc, ntracers, \
                    T, S, GEN, AGE, SED3D, EcoSim, ICM, CoSINE, Feco, \
                    TIMOR, FABM, DVD = f.readline().split()
                f.readline()
                # elements
                ne_local = int(f.readline())
                e_local_to_global = {}
                for i in range(ne_local):
                    local_element_id, global_element_id = f.readline().split()
                    e_local_to_global[local_element_id] = global_element_id
                # points
                np_local = int(f.readline())
                n_local_to_global = {}
                for i in range(np_local):
                    local_node_id, global_node_id = f.readline().split()
                    n_local_to_global[local_node_id] = global_node_id
                # sides
                ns_local = int(f.readline())
                s_local_to_global = {}
                for i in range(ns_local):
                    local_side_id, global_side_id = f.readline().split()
                    s_local_to_global[local_side_id] = global_side_id
                f.readline()  # Header:
                line = f.readline().split()
                if len(line) != 5:
                    line.extend(f.readline().split())
                self.start_year, self.start_month, self.start_day, \
                    self.start_hour, self.utc_start = line
                # nrec, dtout, nspool, nvrt, kz, h0, h_s, h_c, theta_b, \
                #     theta_f, ics = f.readline().split()
                # f.readline()  # (ztot(k),k=1,kz-1),(sigma(k),k=1,nvrt-kz+1)
                # f.readline()  # repeat ne_local, np_local
                _ne_local = None
                _np_local = None
                while _ne_local != ne_local and _np_local != np_local:
                    line = f.readline().split()
                    _np_local = int(float(line[0]))
                    _ne_local = int(float(line[1]))
                for i in range(np_local):
                    x, y, z, flag = map(float, f.readline().split())
                    nodes.setdefault(
                        n_local_to_global[str(i+1)], ((x, y), -z))
                for i in range(ne_local):
                    eids = f.readline().split()[1:]
                    elements.setdefault(
                        e_local_to_global[str(i+1)],
                        list(map(lambda x: n_local_to_global[x], eids)))
            nproc_id = file.name.split('local_to_global_')[-1]
            self.n_local_to_global.setdefault(nproc_id, n_local_to_global)
            self.e_local_to_global.setdefault(nproc_id, e_local_to_global)
            self.s_local_to_global.setdefault(nproc_id, s_local_to_global)
        # sort (not necessary but done for consistency checks)
        nodes = {str(i+1): nodes[str(i+1)] for i in range(len(nodes))}
        elements = {str(i+1): elements[str(i+1)] for i in range(len(elements))}
        self.hgrid = Gr3(nodes=nodes, elements=elements, crs='epsg:4326')
        for vargroup in self.surface_output_vars:
            for vartype in vargroup:
                nc = Dataset(self.filenames[0])
                if vartype.name in nc.variables:
                    setattr(
                        self,
                        vartype.name,
                        OutputVariableCombiner(
                            self,
                            vartype.name,
                            start_step=0
                        ))

    @property
    def n_local_to_global(self):
        if not hasattr(self, '_n_local_to_global'):
            self._n_local_to_global = {}
        return self._n_local_to_global

    @property
    def s_local_to_global(self):
        if not hasattr(self, '_s_local_to_global'):
            self._s_local_to_global = {}
        return self._s_local_to_global

    @property
    def e_local_to_global(self):
        if not hasattr(self, '_e_local_to_global'):
            self._e_local_to_global = {}
        return self._e_local_to_global

    @property
    def stacks(self) -> Dict:
        if not hasattr(self, "_stacks"):
            self._stacks = {}
            for file in self.filenames:
                stack_id = file.name.split("_")[2].split('.')[0]
                nc = Dataset(file)
                self._stacks.setdefault(stack_id, {})
                self._stacks[stack_id].setdefault(
                    'timevector',
                    [self.start_date + timedelta(seconds=float(x))
                     for x in nc['time'][:]]
                    )
        return self._stacks

    @property
    def flattened_timevector(self):
        if not hasattr(self, "_flattened_timevector"):
            flattened_timevector = []
            for stack_id, stack_data in self.stacks.items():
                flattened_timevector.extend(stack_data['timevector'])
            self._flattened_timevector = flattened_timevector
        return self._flattened_timevector

    @property
    def start_date(self):
        if not hasattr(self, "_start_date"):
            start_hour = int(float(self.start_hour))
            self._start_date = datetime(
                int(self.start_year),
                int(self.start_month),
                int(self.start_day),
                start_hour, (start_hour*60) % 60,
                (start_hour*3600) % 60)
            self._start_date -= timedelta(hours=float(self.utc_start))
        return self._start_date


class CombinedOutputs:

    def __init__(self, resource):
        self.dataset = xarray.open_mfdataset(resource)



from datetime import datetime
from time import time
import pathlib
import logging

from pyschism.forcing.source_sink.nwm import NationalWaterModel, NWMElementPairings
from pyschism.mesh import Hgrid

logging.basicConfig(
    format="[%(asctime)s] %(name)s %(levelname)s: %(message)s",
    force=True,
)
logging.captureWarnings(True)

log_level = logging.DEBUG
logging.getLogger('pyschism').setLevel(log_level)

if __name__ == '__main__':

    startdate = datetime(2017, 12, 1)
    rnday = 396
    hgrid = Hgrid.open("./hgrid.gr3", crs="epsg:4326")

    t0 = time()

    #source/sink json files, if not exist, it will call NWMElementPairings to generate.
    sources_pairings = pathlib.Path('./sources.json')
    sinks_pairings = pathlib.Path('./sinks.json')
    output_directory = pathlib.Path('./')

    #input directory which saves downloaded nc files
    #cache = pathlib.Path(f'./{startdate.strftime("%Y%m%d")}')
    #cache.mkdir(exist_ok=True, parents=True)
    cache = pathlib.Path('./NWM_v2.1')

    # check if source/sink json file exists
    if all([sources_pairings.is_file(), sinks_pairings.is_file()]) is False:
        pairings = NWMElementPairings(hgrid)
        sources_pairings.parent.mkdir(exist_ok=True, parents=True)
        pairings.save_json(sources=sources_pairings, sinks=sinks_pairings)
    else:
        pairings = NWMElementPairings.load_json(
            hgrid, 
            sources_pairings, 
            sinks_pairings)

    #check nc files, if not exist will download
    nwm=NationalWaterModel(pairings=pairings, cache=cache)

    nwm.write(output_directory, hgrid, startdate, rnday, overwrite=True)
    print(f'It took {time()-t0} seconds to generate source/sink')



from datetime import datetime
import io
import logging
import os
import pathlib
from enum import IntEnum
from os import PathLike
from typing import Union

from matplotlib.transforms import Bbox
import numpy as numpy
from pandas import DataFrame
from pyproj import CRS, Transformer
from shapely import ops
from shapely.geometry import Point, Polygon
from stormevents.nhc import VortexTrack
import utm

from pyschism.enums import NWSType
from pyschism.forcing.nws.base import NWS
from pyschism.mesh import gridgr3


class HurricaneModel(IntEnum):
    SYMMETRIC = 1
    GAHM = 10

    @classmethod
    def _missing_(cls, value):
        raise ValueError(f"Invalid hurricane model specified: {value}, choose from: {cls.__members__}")


class BestTrackForcing(VortexTrack, NWS):

    def __init__(
        self,
        storm: Union[str, PathLike, DataFrame, io.BytesIO],
        start_date: datetime = None,
        end_date: datetime = None,
        hurricane_model: Union[str, HurricaneModel] = 'gahm',
        windrot: gridgr3.Windrot = None,
        *args,
        **kwargs
    ):

        self.model = hurricane_model
        self.windrot = windrot

        VortexTrack.__init__(
            self,
            storm=storm,
            start_date=start_date,
            end_date=end_date,
            file_deck='b',
            advisories=['BEST'],
        )

    def __str__(self):
        """Returns string used in param.nml"""
        return f"{self.dtype.value}"

    @classmethod
    def from_nhc_bdeck(
        cls,
        nhc_bdeck: PathLike,
        start_date: datetime = None,
        end_date: datetime = None,
    ) -> 'NWS':
        return cls.from_file(path=nhc_bdeck, start_date=start_date, end_date=end_date)

    def summary(
        self, output: Union[str, os.PathLike] = None, overwrite: bool = False,
    ):
        min_storm_speed = numpy.min(self.data['speed'])
        max_storm_speed = numpy.max(self.data['speed'])
        track_length = self.distance
        duration = self.duration
        min_central_pressure = numpy.min(self.data['central_pressure'])
        max_wind_speed = numpy.max(self.data['max_sustained_wind_speed'])
        start_loc = (self.data['longitude'][0], self.data['latitude'][0])
        end_loc = (self.data['longitude'].iloc[-1], self.data['latitude'].iloc[-1])
        f = [
            f'Summary of storm: {self.nhc_code}',
            f'min./max. track speed: {min_storm_speed} m/s, {max_storm_speed} m/s',
            f'min. central pressure: {min_central_pressure} hPa',
            f'max. wind speed: {max_wind_speed} kts',
            f'Starting at: {start_loc} and ended at: {end_loc}',
            f'Total track length: {track_length:.2f} km',
            f'Total track duration: {duration:.2f} days',
        ]
        summary = '\n'.join(f)
        if output is not None:
            if not isinstance(output, pathlib.Path):
                output = pathlib.Path(output)
            if overwrite or not output.exists():
                with open(output, 'w+') as fh:
                    fh.write(summary)
            else:
                logging.debug(f'skipping existing file "{output}"')
        return summary

    def write(
        self,
        path: PathLike,
        overwrite: bool = False,
        windrot: bool = True,
    ):
        VortexTrack.to_file(
            self, path=path/'hurricane-track.dat', overwrite=overwrite)

        # write windrot data
        if windrot is not False and self.windrot is not None:
            windrot = "windrot_geo2proj.gr3" if windrot is True else windrot
            self.windrot.write(path / "windrot_geo2proj.gr3", overwrite)

    @property
    def dtype(self) -> NWSType:
        """Returns the datatype of the object"""
        return NWSType(-1)

    @property
    def model(self) -> HurricaneModel:
        """Return hurricane model used for this best track forcing"""
        return self._model

    @model.setter
    def model(self, value: Union[str, HurricaneModel]):
        """Set hurricane model used for this best track forcing"""
        if isinstance(value, str):
            if value.upper() not in HurricaneModel.__members__:
                raise ValueError(f"Invalid hurricane model specified: {value}, choose from: {HurricaneModel.__members__}")
            value = HurricaneModel[value.upper()]
        elif isinstance(value, int):
            value = HurricaneModel(value)
        elif isinstance(value, HurricaneModel):
            pass
        else:
            raise ValueError(f"Invalid hurricane model specified: {value}")
        self._model = value

    def clip_to_bbox(self, bbox, bbox_crs):
        msg = f'bbox must be a {Bbox} instance.'
        assert isinstance(bbox, Bbox), msg
        bbox_pol = Polygon(
            [
                [bbox.xmin, bbox.ymin],
                [bbox.xmax, bbox.ymin],
                [bbox.xmax, bbox.ymax],
                [bbox.xmin, bbox.ymax],
                [bbox.xmin, bbox.ymin],
            ]
        )
        _switch = True
        unique_dates = numpy.unique(self.data['datetime'])
        _found_start_date = False
        for _datetime in unique_dates:
            records = self.data[self.data['datetime'] == _datetime]
            radii = records['radius_of_last_closed_isobar'].iloc[0]
            radii = 1852.0 * radii  # convert to meters
            lon = records['longitude'].iloc[0]
            lat = records['latitude'].iloc[0]
            _, _, number, letter = utm.from_latlon(lat, lon)
            df_crs = CRS.from_epsg(4326)
            utm_crs = CRS.from_epsg(f'326{number}')
            transformer = Transformer.from_crs(df_crs, utm_crs, always_xy=True)
            p = Point(*transformer.transform(lon, lat))
            pol = p.buffer(radii)
            transformer = Transformer.from_crs(utm_crs, bbox_crs, always_xy=True)
            pol = ops.transform(transformer.transform, pol)
            if _switch is True:
                if not pol.intersects(bbox_pol):
                    continue
                else:
                    self.start_date = records['datetime'].iloc[0]
                    _found_start_date = True
                    _switch = False
                    continue

            else:
                if pol.intersects(bbox_pol):
                    continue
                else:
                    self.end_date = records['datetime'].iloc[0]
                    break

        if _found_start_date is False:
            raise Exception(f'No data within mesh bounding box for storm {self.storm_id}.')

    @property
    def windrot(self):
        return self._windrot

    @windrot.setter
    def windrot(self, windrot: Union[gridgr3.Windrot, None]):
        if not isinstance(windrot, (gridgr3.Windrot, type(None))):
            raise TypeError(
                f"Argument windrot must be of type {gridgr3.Windrot} or None, "
                f"not type {type(windrot)}."
            )
        self._windrot = windrot



from datetime import datetime, timedelta
import logging

from pyschism.mesh import Hgrid, Vgrid
from pyschism.forcing.hycom.hycom2schism import Nudge

'''
outputs:
    SAL_nudge.gr3
    SAL_nu.nc
    TEM_nudge.gr3
    TEM_nu.nc
'''

logging.basicConfig(
    format="[%(asctime)s] %(name)s %(levelname)s: %(message)s",
    force=True,
)
logger = logging.getLogger('pyschism')
logger.setLevel(logging.INFO)

if __name__ == '__main__':


    hgrid=Hgrid.open('./hgrid.gr3', crs='epsg:4326')
    vgrid='./vgrid.in'
    outdir='./'
    start_date = datetime(2018, 1, 1)
    rnday=30

    #ocean_bnd_ids - segment indices, starting from zero.
    nudge=Nudge(hgrid=hgrid, ocean_bnd_ids=[0, 1, 9])

    #rlmax - max relax distance in m or degree
    #rnu_day - max relax strength in days 
    #restart = True will append to the existing nc file, works when first try doesn't break.
    nudge.fetch_data(outdir, vgrid, start_date, rnday, restart=False, rlmax=1, rnu_day=1)



from pyschism.cmd import common as parser_common
from pyschism.cmd.bctides import BctidesCli
from pyschism.cmd.bootstrap import BootstrapCli
from pyschism.cmd.fgrid import FgridCli
from pyschism.cmd.fluxflag import FluxflagCli
from pyschism.cmd.forecast import ForecastCli
from pyschism.cmd.grd2sms import Grd2SmsCli
from pyschism.cmd.hgrid import HgridCli
from pyschism.cmd.outputs import OutputsCli
from pyschism.cmd.sflux import SfluxCli
from pyschism.cmd.sms2grd import Sms2grdCli
from pyschism.cmd.stations import StationsCli
from pyschism.cmd.tvdflag import TvdflagCli
from pyschism.cmd.vgrid import VgridCli

__all__ = [
    "parser_common",
    "BctidesCli",
    "BootstrapCli",
    "FgridCli",
    "FluxflagCli",
    "ForecastCli",
    "Grd2SmsCli",
    "HgridCli",
    "OutputsCli",
    "SfluxCli",
    "Sms2grdCli",
    "StationsCli",
    "TvdflagCli",
    "VgridCli",
]
